
/*  $Header: /cvs/faces/faces/faces/faces.h,v 1.8 2002/02/21 15:42:54 devface Exp $
 *
 *  Contains all the global definitions used by faces.
 *
 *  Copyright (C) 1990-2002  Rich Burridge, Sun Microsystems Inc.
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 *
 *  No responsibility is taken for any errors or inaccuracies inherent
 *  either to the comments or the code of this program, but if
 *  reported to me then an attempt will be made to fix them.
 */

#ifndef _FACES_H
#define _FACES_H

#include <stdlib.h>
#include <string.h>
#include <time.h>

#ifdef  __cplusplus
extern "C" {
#endif

#ifndef FALSE
#define  FALSE  (0 == 1)
#define  TRUE   (0 == 0)
#endif /*FALSE*/

#ifdef DEBUG
#define  IF_DEBUG(stmt)  stmt
#else
#define  IF_DEBUG(stmt)
#endif /*DEBUG*/

#ifndef FACEDIR
#define  FACEDIR  "/usr/local/faces"
#endif /*FACEDIR*/

#define  CLOSE         (void) close       /* To make line happy. */
#define  FCLOSE        (void) fclose
#define  FFLUSH        (void) fflush
#define  FGETS         (void) fgets
#define  FPRINTF       (void) fprintf
#define  FPUTS         (void) fputs
#define  FREE          (void) free
#define  FSEEK         (void) fseek
#define  IOCTL         (void) ioctl
#define  PCLOSE        (void) pclose
#define  POLL          (void) poll
#define  PUTC          (void) putc
#define  SELECT        (void) select
#define  SSCANF        (void) sscanf
#define  SPRINTF       (void) sprintf
#define  STRCAT        (void) strcat
#define  STRCPY        (void) strcpy
#define  STRNCAT       (void) strncat
#define  STRNCPY       (void) strncpy
#define  SYSTEM        (void) system
#define  UNGETC        (void) ungetc
#define  UTIME         (void) utime
#define  WRITE         (void) write

/* Window display types. */
enum disp_type { DISP_NAME,      /* Window type 1. */
                 DISP_OTHER,     /* Window type 2. */
                 DISP_ICON,      /* Icon display. */
                 DISP_BOTH,      /* Both window types. */
                 DISP_ALL        /* Both window types and the icon. */
};

enum dir_type { FORWARD, BACKWARD };   /* Mail folder reading directions. */

/* Different fields in a program record. */
enum field_type { REC_USER,  REC_HOST,
                  REC_LWIN,  REC_RWIN,
                  REC_LICON, REC_RICON };

/* Mail folder types. */
enum folder_type { F_NORMAL, F_IMAP, F_POP };

enum dialog_type { ERROR, WARNING };            /* Message dialog types. */

/* Different types of possible face images. */
enum icon_type { BACKIMAGE, NOFACE,  NOMAIL,   NOPAPER, NOPRINT, 
                 NOTHING,   NOUSERS, ORDINARY, XFACE };

enum image_type { ONSCREEN, OLDOFF, CUROFF };   /* Image types. */

enum just_type { LEFT, RIGHT };                 /* Text justification. */

/* Possible make_iconname() return codes. */
enum min_type { ICON_NOTFOUND,  ICON_MISC_UNKNOWN, ICON_USER,
                ICON_COMMUNITY, ICON_BOTH };

/* Different types of file monitoring performed by this program. */
enum mon_type { MONNEW, MONALL, MONPROG };

/* Property window button types. */
enum pbutton_type { P_OK, P_APPLY, P_CANCEL };

/* X resources used by faces. */
enum res_type { R_WINGEOM, R_ICONGEOM, R_FGCOLOR, R_BGCOLOR,
                R_FONT,    R_BGICON,   R_PERIOD,  R_BELL,
                R_FLASH,   R_LOWER,    R_RAISE,   R_BUT1CLR, R_DISPHOST
#ifdef IMAP
                , R_IMAP
#endif /*IMAP*/
#ifdef POP
                , R_POP
#endif /*POP*/
#ifdef AUDIO_SUPPORT
                , R_AUDIO,   R_AUDIOCMD, R_BELLFILE
#endif /*AUDIO_SUPPORT*/
};

/* Determine order for face type lookup. */
#define  XPMTYPE         1
#define  XBMTYPE         2

/* Keyboard actions. */
#define  KEY_CLEAR       65535   /* Clear faces display      (Del). */
#define  KEY_REFRESH     65293   /* Refresh faces display    (Newline). */
#define  KEY_QUIT        113     /* Terminate faces display  ('q'). */

#define  BLITHEIGHT      48       /* Maximum number of lines in a blit icon. */

#define  CLEARMEM(arg, len)      (void) memset(arg, '\0', len)
#define  GETHOSTNAME(name, len)  (void) gethostname(name, len)

#define  EQUAL(str,val)  !strncmp(str,val,strlen(val))
#define  ICONHEIGHT      64   /* Default height of individual face icons. */
#define  ICONWIDTH       64   /* Default width of individual face icons. */
#define  INC             argc--; argv++;

#ifndef  LINT_CAST
#ifdef   lint
#define  LINT_CAST(arg)  (arg ? 0 : 0)
#else
#define  LINT_CAST(arg)  (arg)
#endif /*lint*/
#endif /*LINT_CAST*/

#define  MAXFIELDS       6    /* Maximum number of fields in program record. */
#define  MAXICONS        9    /* Maximum number of different icon types. */
#define  MAXITYPES       3    /* Maximum number of image types. */

#ifndef  MAXLINE
#define  MAXLINE         200  /* Maximum length for character strings. */
#endif /*MAXLINE*/

#ifndef  MAXPATHS
#define  MAXPATHS        10   /* Maximum number of dirs in FACEPATH list */
#endif /*MAXPATHS*/

#define  MAXTYPES        4    /* Maximum number of different face types. */
#define  NO_PER_ROW      10   /* Default number of faces per row. */

extern char *strlower();

/*  Various programs that are used by faces.
 *  You might need to change these on some machines.
 */

#ifdef AUDIO_SUPPORT
/* Playing sound files:           %s replaced with audio filename. */
#define  AUDIODEF   "play %s" 
#endif /*AUDIO_SUPPORT*/ 

/* Default definitions used by faces. They can be globally changed here. */

#define  MACHINETAB    "machine.tab"   /* Machine aliases table. */
#define  PEOPLETAB     "people.tab"    /* Username aliases table. */
#define  UPDATE_ALIAS  "facemaker"     /* Faces database update alias. */

struct machinfo {              /* Machine/community record. */
    char *machine;             /* Machine name. */
    char *community;           /* Community it belongs to. */
    struct machinfo *next;     /* Pointer to next record. */
};

struct comminfo {              /* Community alias/username records. */
    char *community;           /* Community name. */
    struct peopinfo *people;   /* Chain of alias/usernames. */
    struct comminfo *next;     /* Pointer to next record. */
};

struct peopinfo {              /* Username/alias record. */
    char *alias;               /* Alias for this user. */
    char *community;           /* Real community. */
    char *username;            /* Real username. */
    struct peopinfo *next;     /* Pointer to next record. */
};

struct recinfo {               /* Mail/print information record. */
    char *community;           /* Community name for this person. */
    unsigned char *faceimage;  /* X-Face image or NULL. */
    char *iconname;            /* Name of iconfile for this person. */
    char *username;            /* User name for this person. */
    int total;                 /* Total number of messages/print jobs. */
    int size;                  /* Total size in bytes of print job. */
    int update;                /* If set, the database has been updated. */
    char ts[6];                /* Latest timestamp for this user. */
    struct recinfo *next;      /* Pointer to next record. */
};

enum min_type make_iconname(char *[MAXPATHS+1], char *, char *);

struct recinfo *rec_exists(char *, char *);

char *Malloc(int);
char *get_resource(enum res_type);
char *getname();
char *strlower(char *);

enum icon_type get_icon(char *, enum icon_type);

int get_xbm_icon(char *, enum icon_type);
int get_xpm_icon(char *, enum icon_type);
int init_ws_type(int, char *[]);
int is822header(char *, char *);
int LWSP(int c);
int main(int, char *[]);
int parsefrom(char *, char **, char **);
time_t parsedate(char *s);
int save_resources();

#ifdef AUDIO_SUPPORT
void check_audio(char *);
void play_sound(char *);
#endif /*AUDIO_SUPPORT*/

void a_to_u(char *, char *, char *, char *);
void add_record(char *, char *, char *, int);
void adjust_image(enum disp_type, enum icon_type, int, int);
void create_pixrects(int, int);
void destroy_image();
void display_error(char *);
void display_warning(char *);
void do_check();
void do_key(int);
void draw_screen();
void drop_back();
void garbage_collect();
void get_options(int, char *[]);
void get_xface(char *, unsigned short []);
void h_to_c(char *, char *);
void init_graphics(int *, char *[]);
void load_data(enum icon_type, unsigned char *);
void load_xbm_icon(enum icon_type, unsigned short [256]);
void load_resources();
void make_area(enum disp_type, int, int);
void make_frame(int, char *[]);
void put_resource(enum res_type, char *);
void read_aliases();
void read_machines();
void read_str(char **, char *);
void repl_image(enum disp_type, enum image_type, int, int);
void show_display(int);
void start_tool();
void text(enum disp_type, enum just_type, char *);
void write_cmdline();
void write_resources();

#ifdef  __cplusplus
}
#endif

#endif  /* _FACES_H */
