
/*  $Header: /cvs/faces/faces/faces/get.c,v 1.3 2002/01/15 22:58:46 devface Exp $
 *
 *  Extraction routines used by faces.
 *
 *  Copyright (C) 1990-2002  Rich Burridge, Sun Microsystems Inc.
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 *
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/types.h>
#include <unistd.h>
#include "config.h"
#include "faces.h"
#include "extern.h"

static void getparam(char **s, char *argv[], char *errmes);
static void usage();


/* Add boolean value to saved command line if TRUE. */

static void
cmdbool(int value, char *line, char *arg)
{
    if (value == TRUE) {
        STRCAT(line, arg);
    }
}
 
 
/* Add integer argument to saved command line if not the default value. */
 
static void
cmdint(int value, int defval, char *line, char *arg)
{
    char buf[MAXPATHLEN];
 
    if (value != defval) {
        SPRINTF(buf, arg, value);
        STRCAT(line, buf);
    }
}
 
 
/* Add string argument to saved command line if not NULL. */
 
static void
cmdstr(char *value, char *line, char *arg)
{
    char buf[MAXPATHLEN];
 
    if (value != NULL) {
        SPRINTF(buf, arg, value);
        STRCAT(line, buf);
    }
}


enum icon_type
get_icon(char *dirname, enum icon_type itype)     /* Read in icon file. */
{

/*  Attempts to open the correct face file.
 *
 *  If this is a face.xpm or a face.xbm file, then the XPM or XBM image is
 *  loaded and the appropriate image type is returned, otherwise return an 
 *  icon type of NOFACE.
 */

    char *ptr;

    if ((ptr = strrchr(dirname, '/')) == NULL) {
        return(NOFACE);
    }
    if (EQUAL(ptr+1, "face.xpm")) {
        if (get_xpm_icon(dirname, itype) == 0) {
            return(itype);
        }
    }
    if (EQUAL(ptr+1, "face.xbm")) {
        if (get_xbm_icon(dirname, itype) == 0) {
            return(itype);
        }
    }

    return(NOFACE);
}


/* Read and process command line options. */

void
get_options(int argc, char *argv[])
{
    char *faceenv, *pptr;
    char *next = NULL;      /* The next command line parameter. */
    int i;

    faceenv = getenv("FACEPATH");
    if (faceenv) {
        read_str(&facedirs, faceenv);
    } else {
        facedirs = NULL;
    }

    INC;
    while (argc > 0) {
        if (argv[0][0] == '-') {
            switch (argv[0][1]) {
#ifdef AUDIO_SUPPORT
                case 'A' : do_audio = TRUE;   /* Enable audio support. */
                           break;

                case 'B' : INC;
                           getparam(&bell_sound, argv, 
                                    "-B needs bell audio file");
                           break;

                case 'C' : INC;
                           getparam(&audio_cmd, argv, 
                                    "-C needs audio command");
                           break;
#endif /*AUDIO_SUPPORT*/
                case 'D' : debug = TRUE;
                           break;

#ifdef IMAP
                case 'I' : is_imap = TRUE;
                           break;
#endif /*IMAP*/

                case 'M' : mhflag++;            /* Mail box can shrink. */
                           break;

#ifdef POP
                case 'P' : is_pop = TRUE;
                           break;
#endif /*POP*/

                case 'S' : INC;               /* Alternative spoolfile. */
                           getparam(&next, argv, "-s needs spool directory");
                           SPRINTF(spoolfile, "%s/%s", next, username);
                           break;

                case 'a' : mtype = MONALL;  /* Monitor all of the spoolfile. */
                           break;

                case 'b' : if (argv[0][2] == 'g') {
                               INC;           /* Background color. */
                               getparam(&bgcolor, argv,
                                        "-bg specifies background color");
                               break;
                           }
                           INC;            /* Alternate background pattern. */
                           getparam(&bgicon, argv, "-b needs background icon");
                           break;

                case 'c' : INC;               /* Number of columns. */
                           getparam(&next, argv,
                                    "-c needs number of columns value");
                           maxcols = atoi(next);
                           if (maxcols <= 0 || maxcols > (1152 / ICONWIDTH)) {
                               maxcols = NO_PER_ROW;
                           }
                           break;

                case 'd' :                    /* X11 display information. */
                           INC;
                           getparam(&display, argv, 
                                    "-d needs display information");
                           break;

                case 'e' : mtype = MONPROG;   /* User specified program. */
                           INC;
                           getparam(&userprog, argv, "-e needs user program");
                           break;

                case 'f' : if (argv[0][2] == 'g') {
                               INC;           /* Background color. */
                               getparam(&fgcolor, argv,
                                        "-fg specifies foreground color");
                               break;
                           } else if (argv[0][2] == 'n') {
                               INC;           /* font name. */
                               getparam(&fontname, argv,
                                        "-fn specifies font name");
                               break;
                           }
                           INC;             /* New directory for face icons. */
                           getparam(&facedirs, argv,
                                    "-f needs face directory path");
                           break;

                case 'g' : INC;               /* X11 geometry information. */
                           getparam(&geometry, argv,
                                    "-g needs geometry information");
                           posspec = 1;
                           break;

                case 'h' : INC;               /* Face image height. */
                           getparam(&next, argv, "-h needs height value");
                           imageheight = atoi(next);
                           if (imageheight < ICONHEIGHT || imageheight > 250) {
                               FPRINTF(stderr, 
                                       "Illegal -h value; resetting.\n");
                               imageheight = ICONHEIGHT;
                           }
                           break;

                case 'i' : if (!strncmp(&argv[0][1], "iconic", 6)) {
                               iconic = 1;
                           } else if (argv[0][2] == '\0') {
                               invert = 1;
                           }
                           break;

                case 'l' : INC;           /* Faces window title line label. */
                           getparam(&faces_label, argv, 
                                    "-l needs window label");
                           break;

                case 'n' : dontshowno = 1;  /* Don't show number of messages. */
                           break;

                case 'p' : INC;             /* No. of seconds between checks. */
                           getparam(&next, argv, "-p needs period time");
                           period = atoi(next);
                           break;

                case 'r' : INC;               /* Same as -i, but for X11 */
                           invert = 1;
                           break;

                case 's' : INC;               /* Alternative spoolfile. */
                           getparam(&spoolfile, argv, "-s needs spool file");
                           break;

                case 't' : dontshowtime = 1;  /* Do not show timestamps. */
                           break;

                case 'u' : dontshowuser = 1;  /* Do not show usernames. */
                           break;

                case '?' :
                case 'v' : usage();
                           break;

                case 'w' : INC;               /* Face image width. */
                           getparam(&next, argv, "-w needs width value");
                           imagewidth = atoi(next);
                           if (imagewidth < ICONWIDTH || imagewidth > 250) {
                               FPRINTF(stderr, 
                                       "Illegal -w value; resetting.\n");
                               imagewidth = ICONWIDTH;
                           }
                           break; 

                default  : usage();
                           exit(1);
            }
        }
        INC;
    }

/* Parse the face path from the parameter line... */

    for (i = 0, pptr = facedirs; pptr && (i <= MAXPATHS); i++) {
        if ((*pptr == '\0') || (*pptr == ':')) {

/* Null entry in face path means insert compiled-in default. */

            facepath[i] = FACEDIR;
            if (*(pptr++) == '\0') {
                break;
            }
        } else {
            facepath[i] = pptr;
            pptr = strchr(pptr, ':');
            if (pptr) {
                *(pptr++) = '\0';
            }
        }
    }
}


void
get_xface(char *ibuf, unsigned short obuf[])   /* Extract "raw" X-Face data. */
{
    char *ptr;
    int i, j, temp;

    ptr = ibuf;
    for (i = 0; i < BLITHEIGHT; i++) {
#ifdef REVORDER
        for (j = 2; j >= 0; j--)
#else
        for (j = 0; j < 3; j++)
#endif /*REVORDER*/
        {
            while (*ptr == ' ' || *ptr == '\t' || *ptr == '\n') {
                ptr++;
            }
            SSCANF(ptr, "0x%X", &temp);
            obuf[i*4 + j] = (short) temp;
            if ((ptr = strchr(ptr, ',')) == NULL) {
                return;
            }
            ptr++;
        }
        obuf[i*4 + 3] = 0;
    }
    for (i = BLITHEIGHT; i < iconheight; i++) {
        for (j = 0; j < 4; j++) {
            obuf[i*4 + j] = 0;
        }
    }
}


char *
getname()       /* Get users name from passwd entry. */
{
    char *username;
    struct passwd *pwent;

    if (!(username = getlogin())) {
        pwent = getpwuid(getuid());
        username = pwent->pw_name;
        endpwent();                   /* Close the passwd file */
    }

    return(username);
}


static void
getparam(char **s, char *argv[], char *errmes)
{
    if (*argv != NULL) {
        read_str(s, *argv);
    } else {
        FPRINTF(stderr,"%s: %s as next argument.\n", progname, errmes);
        exit(1);
    }
}


char *
strlower(char *str)
{
    char *c;

    for (c = str; *c != '\0'; c++) {
        if (isupper((int) *c)) {
            *c = tolower(*c);
        }
    }

    return(str);
}


static void
usage()      /* Print faces usage message. */
{
    FPRINTF(stderr, "%s version %s\n\n", progname, VERSION);
    FPRINTF(stderr, "Usage: %s ", progname);
#ifdef AUDIO_SUPPORT
    FPRINTF(stderr, "\t[-A] [-B bell_sound_file] [-C audio_cmd]\n");
#endif /*AUDIO_SUPPORT*/

    FPRINTF(stderr, "\t");
#ifdef IMAP
    FPRINTF(stderr, "[-I] ");
#endif /*IMAP*/
    FPRINTF(stderr, "[-M] ");
#ifdef POP
    FPRINTF(stderr, "[-P] ");
#endif /*POP*/
    FPRINTF(stderr, "[-S spooldir] [-a] [-b background]\n");
    FPRINTF(stderr, "\t[-bg background_color] [-c columns]\n");
    FPRINTF(stderr, "\t[-d display] [-e progname] [-f facepath]\n");
    FPRINTF(stderr, "\t[-fg foreground_color] [-fn fontname]\n");
    FPRINTF(stderr, "\t[-g geometry] [-h height]\n");
    FPRINTF(stderr, "\t[-i] [-iconic] [-l label] [-n] [-p period] [-rv]\n");
    FPRINTF(stderr, "\t[-s spoolfile] [-t] [-u] [-v] [-w width] [-?]\n\n");
    FPRINTF(stderr, "Keyboard accelerators:\n");
    FPRINTF(stderr, "\t^L  - clear faces display.\n");
    FPRINTF(stderr, "\tdel - refresh faces display.\n");
    FPRINTF(stderr, "\tq   - terminate faces display.\n");
    exit(1);
}


void
write_cmdline()
{
    char line[MAXPATHLEN];

    line[0] = '\0';

#ifdef AUDIO_SUPPORT
    cmdbool(do_audio,  line, " -A ");           /* Enable audio support. */
    cmdstr(bell_sound, line, " -B %s ");        /* Bell audio file. */
    cmdstr(audio_cmd,  line, " -C %s ");        /* Audio command */
#endif /*AUDIO_SUPPORT*/

    cmdbool(debug,        line, " -D ");        /* Turn on debugging. */
    cmdbool(mhflag,       line, " -M ");        /* Mail box can shrink. */
    cmdbool(iconic,       line, " -iconic ");   /* Start faces iconic. */
    cmdbool(invert,       line, " -i ");        /* Invert display. */
    cmdbool(dontshowno,   line, " -n ");        /* Don't show # of messages. */
    cmdbool(dontshowtime, line, " -t ");        /* Don't show timestamps. */
    cmdbool(dontshowuser, line, " -u ");        /* Don't show usernames. */

    if (mtype == MONALL) {                       /* Monitor all of spoolfile. */
        STRCAT(line, " -a ");
    }
    if (mtype == MONPROG) {                      /* User specified program. */
        cmdstr(userprog,  line, " -e %s ");
    }

    cmdint(imageheight, ICONHEIGHT, line, " -h %d ");    /* Image height. */ 
    cmdint(imagewidth,  ICONWIDTH,  line, " -w %d ");    /* Image width. */
    cmdint(maxcols,     NO_PER_ROW, line, " -c %1d ");   /* # of columns. */
    cmdint(period,      60,         line, " -p %d ");    /* Period length. */

    cmdstr(bgcolor,     line, " -bg %s ");      /* Background color. */
    cmdstr(bgicon,      line, " -b %s ");       /* Background pattern. */
    cmdstr(display,     line, " -d %s ");       /* X11 display information. */
    cmdstr(fgcolor,     line, " -fg %s ");      /* Foreground color. */
    cmdstr(fontname,    line, " -fn %s ");      /* Alternate fontname. */
    cmdstr(facedirs,    line, " -f %s ");       /* List of face directories. */
    cmdstr(geometry,    line, " -g %s ");       /* X11 geometry information. */
    cmdstr(faces_label, line, " -l %s ");       /* Window title line. */
    cmdstr(spoolfile,   line, " -s %s ");       /* Spoolfile to monitor. */
}
