
/*  $Header: /cvs/faces/faces/faces/gtk.c,v 1.18 2002/04/08 15:14:18 devface Exp $
 *
 *  X11 dependent graphics routines used by faces,
 *  the visual mail and print job monitor.
 * 
 *  Copyright (C) 1990-2002  Rich Burridge, Sun Microsystems Inc.
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 * 
 *  No responsibility is taken for any errors or inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 *
 */
 
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "faces.h"
#include "extern.h"
#include "nomail.xbm"
#include "dialog_error.xpm"
#include "dialog_warning.xpm"

#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <X11/Xresource.h>

#include "textutil.h"

#define  DEF_FONT            "fixed"
#define  FONT                "-*-helvetica-*-r-*--10-*"

#define  FACES_BORDER_WIDTH  2
#define  FRAME_MASK          (GDK_BUTTON_PRESS_MASK | GDK_EXPOSURE_MASK | \
                              GDK_BUTTON_MOTION_MASK | GDK_KEY_PRESS_MASK)
#define  ICON_MASK           GDK_EXPOSURE_MASK

static GtkWidget *window;
static GtkWidget *canvas;
static GdkBitmap *iconbm;

static GtkWidget *error_label, *warning_label;
static GtkWidget *error_dialog, *errorpm, *warning_dialog, *warningpm;

static GtkWidget *create_dialog(enum dialog_type, char *, GtkWidget **);
static void do_error_dialog_ok(GtkButton *, gpointer);
static void do_warning_dialog_ok(GtkButton *, gpointer);

static GtkWidget *props_window = NULL;
static GtkWidget *name_label;
static GtkWidget *type_hbox;
static GtkWidget *mail_rb; 
static GtkWidget *custom_rb;
static GtkWidget *normal_rb;
static GtkWidget *imap_rb;
static GtkWidget *pop_rb;
static GtkWidget *interval_value; 
static GtkWidget *messages_check; 
static GtkWidget *time_check; 
static GtkWidget *user_check; 
static GtkWidget *name;

static gboolean dismiss_props(GtkWidget *, GdkEvent *, gpointer);
static void mail_toggled(GtkToggleButton *, gpointer);
static void custom_toggled(GtkToggleButton *, gpointer);
static void ok_clicked(GtkButton *, gpointer);
static void apply_clicked(GtkButton *, gpointer);
static void cancel_clicked(GtkButton *, gpointer);
static void set_mail_custom_widgets(gboolean);
static void do_props_button_clicked(enum pbutton_type);
static void set_props_state();

#ifdef IMAP
static GtkWidget *imap_popup;
static GtkWidget *imap_server;
static GtkWidget *imap_login;
static GtkWidget *imap_password;

static void do_imap_login(GtkButton *, gpointer);
static void do_imap_cancel(GtkButton *, gpointer);
#endif /*IMAP*/

#ifdef POP
static GtkWidget *pop_popup;
static GtkWidget *pop_server;
static GtkWidget *pop_login;
static GtkWidget *pop_password;

static void do_pop_login(GtkButton *, gpointer);
static void do_pop_cancel(GtkButton *, gpointer);
#endif /*POP*/

static GdkGC *gc, *tilegc, *xpmgc;
static GdkWindow *canvas_win;
static GdkPixmap *images[MAXICONS] = { /* Array of the different icon images. */
    0, 0, 0, 0, 0, 0, 0, 0, 0
};
static GdkPixmap *pr[3] = {            /* Current memory Pixmaps. */
    0, 0, 0 
};
static GdkPixmap *old_pr[2] = {        /* Previous memory Pixmaps. */
    0, 0 
};

static Display *dpy;
static GdkAtom protocol_atom, kill_atom;
static GdkColormap *cmap;
static GdkFont *sfont;
static GdkGCValues gc_val;
static GdkGCValuesMask gc_mask;
static XrmDatabase faces_DB;           /* Combined resources database. */
static XSizeHints size;
static int error_displayed = FALSE;
static int warning_displayed = FALSE;
static int depth;
static GdkColor backgnd, foregnd;
static guint timerid;
static guint login_timerid = 0;

static void do_pause(void);
static gboolean button_press(GtkWidget *, GdkEventButton *, gpointer);
static gboolean canvas_resize(GtkWidget *, GdkEventConfigure *, gpointer);
static gboolean dismiss_faces(GtkWidget *, GdkEvent *, gpointer);
static gboolean key_press(GtkWidget *, GdkEventKey *, gpointer );
static gboolean process_expose(GtkWidget *, GdkEventExpose *, gpointer);
static gint display_login_popup(gpointer data);
static gint update_timer(gpointer data);


/* Put new face in memory pixrect. */

void
adjust_image(enum disp_type dtype, enum icon_type itype, int row, int column)
{
    if (mtype == MONNEW) {
        gdk_window_copy_area(pr[(int) dtype], gc, imagewidth, 0,
                             pr[(int) dtype], 0, 0,
                             (maxcols-1)*imagewidth, imageheight);
    }

    gdk_gc_set_function(gc, GDK_SET);
    gdk_window_copy_area(pr[(int) dtype], gc, 
                         column*imagewidth, row*imagewidth,
                         pr[(int) dtype], column*imagewidth, row*imagewidth,
                         imagewidth, imageheight);

    gdk_gc_set_function(gc, GDK_COPY);
    gdk_window_copy_area(pr[(int) dtype], gc,
                         column*imagewidth, row*imageheight,
                         images[(int) itype], 0, 0, iconwidth, iconwidth);
}

 
static void
apply_clicked(GtkButton *button, gpointer user_data)
{
    do_props_button_clicked(P_APPLY);
}



static void
beep_flash(int beeps, int flashes)    /* Perform visual feedback. */
{
#ifdef AUDIO_SUPPORT
    if (do_audio && bell_sound != NULL) {
        play_sound(bell_sound);
    } else
#endif /*AUDIO_SUPPORT*/
    {
        while (beeps > 0) {
            gdk_beep();
            do_pause();
            beeps--;
        }
    }

    while (flashes > 0) {
        gdk_gc_set_function(gc, GDK_COPY_INVERT);
        gdk_window_copy_area(canvas_win, gc, 0, 0, canvas_win, 0, 0,
                             (unsigned int) width, (unsigned int) height);
        do_pause();

        gdk_window_copy_area(canvas_win, gc, 0, 0, canvas_win, 0, 0, 
                             (unsigned int) width, (unsigned int) height);
        do_pause();
 
        gdk_window_copy_area(pr[(int) DISP_ICON], gc, 0, 0,
                             pr[(int) DISP_ICON], 0, 0,
                             (unsigned int) width, (unsigned int) height);
        gdk_window_set_icon(window->window, NULL, pr[(int) DISP_ICON], iconbm);
        do_pause();

        gdk_window_copy_area(pr[(int) DISP_ICON], gc, 0, 0,
                             pr[(int) DISP_ICON], 0, 0,
                             (unsigned int) width, (unsigned int) height);
        gdk_window_set_icon(window->window, NULL, pr[(int) DISP_ICON], iconbm);
        do_pause();

        flashes--;
    }   
    gdk_gc_set_function(gc, GDK_COPY);
}


static void
cancel_clicked(GtkButton *button, gpointer user_data)
{
    do_props_button_clicked(P_CANCEL);
}


static gboolean
canvas_resize(GtkWidget *widget, GdkEventConfigure *event, gpointer user_data)
{
    canvas->requisition.width = width;
    canvas->requisition.height = height;

    return(TRUE);
}


static GtkWidget *
create_dialog(enum dialog_type dtype, char *message, GtkWidget **label)
{
    GtkWidget *dialog, *vbox, *hbox, *pixmap = NULL;
    GtkWidget *action_area, *ok_button;

    dialog = gtk_dialog_new();
    gtk_object_set_data(GTK_OBJECT(dialog), "dialog", dialog);

    switch (dtype) {
        case ERROR:
            gtk_window_set_title(GTK_WINDOW(dialog), _("Error"));
            break;

        case WARNING:
            gtk_window_set_title(GTK_WINDOW(dialog), _("Warning"));
            break;
    }

    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_window_set_policy(GTK_WINDOW(dialog), TRUE, TRUE, FALSE);

    vbox = GTK_DIALOG(dialog)->vbox;
    gtk_object_set_data(GTK_OBJECT(dialog), "vbox", vbox);
    gtk_widget_show(vbox);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_ref(hbox);
    gtk_object_set_data_full(GTK_OBJECT(dialog), "hbox", hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(hbox);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    switch (dtype) {
        case ERROR:
            pixmap = errorpm;
            break;

        case WARNING:
            pixmap = warningpm;
            break;
    }

    gtk_widget_ref(pixmap);
    gtk_object_set_data_full(GTK_OBJECT(dialog), "pixmap", pixmap,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(pixmap);
    gtk_box_pack_start(GTK_BOX(hbox), pixmap, TRUE, TRUE, 0);

    *label = gtk_label_new(_(message));
    gtk_widget_ref(*label);
    gtk_object_set_data_full(GTK_OBJECT(dialog), "label", *label,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(*label);
    gtk_box_pack_start(GTK_BOX(hbox), *label, FALSE, FALSE, 0);

    action_area = GTK_DIALOG(dialog)->action_area;
    gtk_object_set_data(GTK_OBJECT(dialog), "action_area", action_area);
    gtk_widget_show(action_area);
    gtk_container_set_border_width(GTK_CONTAINER(action_area), 10);
 
    ok_button = gtk_button_new_with_label(_("OK"));
    gtk_widget_ref(ok_button);
    gtk_object_set_data_full(GTK_OBJECT(dialog), "ok_button", ok_button,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(ok_button);
    gtk_box_pack_start(GTK_BOX(action_area), ok_button, FALSE, FALSE, 0);

    switch (dtype) {
        case ERROR:
            gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
                       GTK_SIGNAL_FUNC(do_error_dialog_ok),
                       (gpointer) dialog);
            break;

        case WARNING:
            gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
                       GTK_SIGNAL_FUNC(do_warning_dialog_ok),
                       (gpointer) dialog);
            break;
    }

    return(dialog);
}


#ifdef IMAP
GtkWidget*
create_imap_popup(void)
{
    GtkWidget *imap_popup;
    GtkWidget *table;
    GtkWidget *login_button;
    GtkWidget *cancel_button;
    GtkWidget *server_label;
    GtkWidget *login_label;
    GtkWidget *password_label;

    imap_popup = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_object_set_data(GTK_OBJECT(imap_popup), "imap_popup", imap_popup);
    gtk_window_set_title(GTK_WINDOW(imap_popup), _("IMAP Information"));
    gtk_window_set_position(GTK_WINDOW(imap_popup), GTK_WIN_POS_CENTER);
    gtk_window_set_modal(GTK_WINDOW(imap_popup), TRUE);
    gtk_window_set_policy(GTK_WINDOW(imap_popup), TRUE, TRUE, FALSE);

    table = gtk_table_new(4, 2, FALSE);
    gtk_widget_ref(table);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "table", table,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(table);
    gtk_container_add(GTK_CONTAINER(imap_popup), table);
    gtk_container_set_border_width(GTK_CONTAINER(table), 5);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);
 
    imap_server = gtk_entry_new();
    gtk_widget_ref(imap_server);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "imap_server", 
                             imap_server, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(imap_server);
    gtk_table_attach(GTK_TABLE(table), imap_server, 1, 2, 0, 1,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
 
    imap_login = gtk_entry_new();
    gtk_widget_ref(imap_login);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "imap_login", imap_login,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(imap_login);
    gtk_table_attach(GTK_TABLE(table), imap_login, 1, 2, 1, 2,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0);
 
    imap_password = gtk_entry_new();
    gtk_widget_ref(imap_password);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "imap_password", 
                           imap_password, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(imap_password);
    gtk_table_attach(GTK_TABLE(table), imap_password, 1, 2, 2, 3,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_entry_set_visibility(GTK_ENTRY(imap_password), FALSE);
 
    login_button = gtk_button_new_with_label(_("Login"));
    gtk_widget_ref(login_button);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "login_button", 
                            login_button, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(login_button);
    gtk_table_attach(GTK_TABLE(table), login_button, 0, 1, 3, 4,
                     (GtkAttachOptions) (0),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_widget_set_usize(login_button, 64, -2);
 
    cancel_button = gtk_button_new_with_label(_("Cancel"));
    gtk_widget_ref(cancel_button);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "cancel_button", 
                           cancel_button, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(cancel_button);
    gtk_table_attach(GTK_TABLE(table), cancel_button, 1, 2, 3, 4,
                     (GtkAttachOptions) (0),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_widget_set_usize(cancel_button, 64, -2);

    server_label = gtk_label_new(_("IMAP Server:"));
    gtk_widget_ref(server_label);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "server_label", 
                            server_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(server_label);
    gtk_table_attach(GTK_TABLE(table), server_label, 0, 1, 0, 1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(server_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(server_label), 0, 0.5);

    login_label = gtk_label_new(_("Login:"));
    gtk_widget_ref(login_label);
    gtk_object_set_data_full(GTK_OBJECT (imap_popup), "login_label", 
                             login_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(login_label);
    gtk_table_attach(GTK_TABLE(table), login_label, 0, 1, 1, 2,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(login_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(login_label), 0, 0.5);

    password_label = gtk_label_new(_("Password:"));
    gtk_widget_ref(password_label);
    gtk_object_set_data_full(GTK_OBJECT(imap_popup), "password_label", 
                          password_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(password_label);
    gtk_table_attach(GTK_TABLE(table), password_label, 0, 1, 2, 3,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(password_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(password_label), 0, 0.5);

    gtk_signal_connect(GTK_OBJECT(login_button), "clicked",
                       GTK_SIGNAL_FUNC(do_imap_login),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
                       GTK_SIGNAL_FUNC(do_imap_cancel),
                       NULL);

    return(imap_popup);
}
#endif /*IMAP*/


/* Create pixrects for the face images. */

void
create_pixrects(int width, int height)
{
    old_pr[(int) DISP_NAME]  = pr[(int) DISP_NAME];
    old_pr[(int) DISP_OTHER] = pr[(int) DISP_OTHER];

    make_area(DISP_NAME, width, height);
    repl_image(DISP_NAME, CUROFF, width, height);

    if (toclear && mtype == MONNEW) {
        repl_image(DISP_OTHER, OLDOFF, width, height);
        toclear = 0;
    }

    make_area(DISP_OTHER, width, height);
    repl_image(DISP_OTHER, CUROFF, width, height);
 
    if (mtype == MONNEW && old_pr[(int) DISP_NAME] != 0) {
        gdk_window_copy_area(pr[(int) DISP_NAME], gc, 0, 0,
                             old_pr[(int) DISP_NAME], 0, 0,
                             (unsigned int) width, (unsigned int) height);
    }
    if (mtype == MONNEW && old_pr[(int) DISP_OTHER] != 0) {
        gdk_window_copy_area(pr[(int) DISP_OTHER], gc, 0, 0,
                             old_pr[(int) DISP_OTHER], 0, 0,
                             (unsigned int) width, (unsigned int) height);
    }
}


#ifdef POP
GtkWidget*
create_pop_popup(void)
{
    GtkWidget *pop_popup;
    GtkWidget *table;
    GtkWidget *login_button;
    GtkWidget *cancel_button;
    GtkWidget *server_label;
    GtkWidget *login_label;
    GtkWidget *password_label;

    pop_popup = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_object_set_data(GTK_OBJECT(pop_popup), "pop_popup", pop_popup);
    gtk_window_set_title(GTK_WINDOW(pop_popup), _("POP Information"));
    gtk_window_set_position(GTK_WINDOW(pop_popup), GTK_WIN_POS_CENTER);
    gtk_window_set_modal(GTK_WINDOW(pop_popup), TRUE);
    gtk_window_set_policy(GTK_WINDOW(pop_popup), TRUE, TRUE, FALSE);

    table = gtk_table_new(4, 2, FALSE);
    gtk_widget_ref(table);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "table", table,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(table);
    gtk_container_add(GTK_CONTAINER(pop_popup), table);
    gtk_container_set_border_width(GTK_CONTAINER(table), 5);
    gtk_table_set_row_spacings(GTK_TABLE(table), 5);
    gtk_table_set_col_spacings(GTK_TABLE(table), 5);

    pop_server = gtk_entry_new();
    gtk_widget_ref(pop_server);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "pop_server",
                             pop_server, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(pop_server);
    gtk_table_attach(GTK_TABLE(table), pop_server, 1, 2, 0, 1,
                      (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                      (GtkAttachOptions) (0), 0, 0);
        
    pop_login = gtk_entry_new();
    gtk_widget_ref(pop_login);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "pop_login", pop_login,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(pop_login);
    gtk_table_attach(GTK_TABLE(table), pop_login, 1, 2, 1, 2,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0);
        
    pop_password = gtk_entry_new();
    gtk_widget_ref(pop_password);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "pop_password",
                           pop_password, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(pop_password);
    gtk_table_attach(GTK_TABLE(table), pop_password, 1, 2, 2, 3,
                     (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_entry_set_visibility(GTK_ENTRY(pop_password), FALSE);

    login_button = gtk_button_new_with_label(_("Login"));
    gtk_widget_ref(login_button);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "login_button",
                            login_button, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(login_button);
    gtk_table_attach(GTK_TABLE(table), login_button, 0, 1, 3, 4,
                     (GtkAttachOptions) (0),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_widget_set_usize(login_button, 64, -2);

    cancel_button = gtk_button_new_with_label(_("Cancel"));
    gtk_widget_ref(cancel_button);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "cancel_button",
                           cancel_button, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(cancel_button);
    gtk_table_attach(GTK_TABLE(table), cancel_button, 1, 2, 3, 4,
                     (GtkAttachOptions) (0),
                     (GtkAttachOptions) (0), 0, 0);
    gtk_widget_set_usize(cancel_button, 64, -2);

    server_label = gtk_label_new(_("POP Server:"));
    gtk_widget_ref(server_label);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "server_label",
                            server_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(server_label);
    gtk_table_attach(GTK_TABLE(table), server_label, 0, 1, 0, 1,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(server_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(server_label), 0, 0.5);

    login_label = gtk_label_new(_("Login:"));
    gtk_widget_ref(login_label);
    gtk_object_set_data_full(GTK_OBJECT (pop_popup), "login_label",
                             login_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(login_label);
    gtk_table_attach(GTK_TABLE(table), login_label, 0, 1, 1, 2,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(login_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(login_label), 0, 0.5);

    password_label = gtk_label_new(_("Password:"));
    gtk_widget_ref(password_label);
    gtk_object_set_data_full(GTK_OBJECT(pop_popup), "password_label",
                          password_label, (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(password_label);
    gtk_table_attach(GTK_TABLE(table), password_label, 0, 1, 2, 3,
                     (GtkAttachOptions) (GTK_FILL),
                     (GtkAttachOptions) (0), 5, 5);
    gtk_label_set_justify(GTK_LABEL(password_label), GTK_JUSTIFY_RIGHT);
    gtk_misc_set_alignment(GTK_MISC(password_label), 0, 0.5);

    gtk_signal_connect(GTK_OBJECT(login_button), "clicked",
                       GTK_SIGNAL_FUNC(do_pop_login),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(cancel_button), "clicked",
                       GTK_SIGNAL_FUNC(do_pop_cancel),
                       NULL);

    return(pop_popup);
}
#endif /*POP*/


static GtkWidget *
create_properties(void)
{
    GtkWidget *properties;
    GtkWidget *props_vbo;
    GtkWidget *mon_hbox;
    GtkWidget *mon_label;
    GSList *mon_hbox_group = NULL;
    GtkWidget *type_label;
    GSList *type_hbox_group = NULL;
    GtkWidget *name_hbox;
    GtkWidget *int_hbox;
    GtkWidget *int_label;
    GtkObject *interval_value_adj;
    GtkWidget *int_secs;
    GtkWidget *button_hbox;
    GtkWidget *ok_button;
    GtkWidget *apply_button;
    GtkWidget *cancel_button;

    properties = gtk_window_new(GTK_WINDOW_POPUP);
    gtk_object_set_data(GTK_OBJECT(properties), "properties", properties);
    gtk_window_set_title(GTK_WINDOW(properties), _("Properties"));

    props_vbo = gtk_vbox_new(FALSE, 0);
    gtk_widget_ref(props_vbo);
    gtk_object_set_data_full(GTK_OBJECT(properties), "props_vbo", props_vbo,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(props_vbo);
    gtk_container_add(GTK_CONTAINER(properties), props_vbo);

    mon_hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_ref(mon_hbox);
    gtk_object_set_data_full(GTK_OBJECT(properties), "mon_hbox", mon_hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(mon_hbox);
    gtk_box_pack_start(GTK_BOX(props_vbo), mon_hbox, TRUE, TRUE, 0);

    mon_label = gtk_label_new(_("Monitor:"));
    gtk_widget_ref(mon_label);
    gtk_object_set_data_full(GTK_OBJECT(properties), "mon_label", mon_label,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(mon_label);
    gtk_box_pack_start(GTK_BOX(mon_hbox), mon_label, FALSE, FALSE, 0);
    gtk_misc_set_padding(GTK_MISC(mon_label), 5, 0);

    mail_rb = gtk_radio_button_new_with_label(mon_hbox_group, _("Mail"));
    mon_hbox_group = gtk_radio_button_group(GTK_RADIO_BUTTON(mail_rb));
    gtk_widget_ref(mail_rb);
    gtk_object_set_data_full(GTK_OBJECT(properties), "mail_rb", mail_rb,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(mail_rb);
    gtk_box_pack_start(GTK_BOX(mon_hbox), mail_rb, FALSE, FALSE, 0);

    custom_rb = gtk_radio_button_new_with_label(mon_hbox_group, _("Custom"));
    mon_hbox_group = gtk_radio_button_group(GTK_RADIO_BUTTON (custom_rb));
    gtk_widget_ref(custom_rb);
    gtk_object_set_data_full(GTK_OBJECT(properties), "custom_rb", custom_rb,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(custom_rb);
    gtk_box_pack_start(GTK_BOX(mon_hbox), custom_rb, FALSE, FALSE, 0);

    type_hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_ref(type_hbox);
    gtk_object_set_data_full(GTK_OBJECT(properties), "type_hbox", type_hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(type_hbox);
    gtk_box_pack_start(GTK_BOX(props_vbo), type_hbox, TRUE, TRUE, 0);

    type_label = gtk_label_new(_("Folder type:"));
    gtk_widget_ref(type_label);
    gtk_object_set_data_full(GTK_OBJECT(properties), "type_label", type_label,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(type_label);
    gtk_box_pack_start(GTK_BOX(type_hbox), type_label, FALSE, FALSE, 0);
    gtk_misc_set_padding(GTK_MISC(type_label), 5, 0);

    normal_rb = gtk_radio_button_new_with_label(type_hbox_group, _("Normal"));
    type_hbox_group = gtk_radio_button_group(GTK_RADIO_BUTTON(normal_rb));
    gtk_widget_ref(normal_rb);
    gtk_object_set_data_full(GTK_OBJECT(properties), "normal_rb", normal_rb,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(normal_rb);
    gtk_box_pack_start(GTK_BOX(type_hbox), normal_rb, FALSE, FALSE, 0);

    imap_rb = gtk_radio_button_new_with_label(type_hbox_group, _("IMAP"));
    type_hbox_group = gtk_radio_button_group(GTK_RADIO_BUTTON (imap_rb));
    gtk_widget_ref(imap_rb);
    gtk_object_set_data_full(GTK_OBJECT(properties), "imap_rb", imap_rb,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(imap_rb);
    gtk_box_pack_start(GTK_BOX(type_hbox), imap_rb, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(imap_rb, (is_imap == 1) ? TRUE : FALSE);

    pop_rb = gtk_radio_button_new_with_label(type_hbox_group, _("POP"));
    type_hbox_group = gtk_radio_button_group(GTK_RADIO_BUTTON(pop_rb));
    gtk_widget_ref(pop_rb);
    gtk_object_set_data_full(GTK_OBJECT(properties), "pop_rb", pop_rb,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(pop_rb);
    gtk_box_pack_start(GTK_BOX(type_hbox), pop_rb, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(pop_rb, (is_pop == 1) ? TRUE : FALSE);

    name_hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_ref(name_hbox);
    gtk_object_set_data_full(GTK_OBJECT(properties), "name_hbox", name_hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(name_hbox);
    gtk_box_pack_start(GTK_BOX(props_vbo), name_hbox, TRUE, TRUE, 0);

    name_label = gtk_label_new(_("Folder name:"));
    gtk_widget_ref(name_label);
    gtk_object_set_data_full(GTK_OBJECT(properties), "name_label", name_label,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(name_label);
    gtk_box_pack_start(GTK_BOX(name_hbox), name_label, FALSE, FALSE, 0);
    gtk_misc_set_padding(GTK_MISC(name_label), 5, 0);

    name = gtk_entry_new();
    gtk_widget_ref(name);
    gtk_object_set_data_full(GTK_OBJECT(properties), "name", name,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(name);
    gtk_box_pack_start(GTK_BOX(name_hbox), name, TRUE, TRUE, 0);

    int_hbox = gtk_hbox_new(FALSE, 0);
    gtk_widget_ref(int_hbox);
    gtk_object_set_data_full(GTK_OBJECT(properties), "int_hbox", int_hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(int_hbox);
    gtk_box_pack_start(GTK_BOX(props_vbo), int_hbox, TRUE, TRUE, 0);

    int_label = gtk_label_new(_("Monitoring interval:"));
    gtk_widget_ref(int_label);
    gtk_object_set_data_full(GTK_OBJECT(properties), "int_label", int_label,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(int_label);
    gtk_box_pack_start(GTK_BOX(int_hbox), int_label, FALSE, FALSE, 0);
    gtk_misc_set_padding(GTK_MISC(int_label), 5, 0);

    interval_value_adj = gtk_adjustment_new(60, 0, 100, 1, 10, 10);
    interval_value = gtk_spin_button_new(GTK_ADJUSTMENT(interval_value_adj),
                                         1, 0);
    gtk_widget_ref(interval_value);
    gtk_object_set_data_full(GTK_OBJECT(properties), "interval_value", 
                             interval_value,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(interval_value);
    gtk_box_pack_start(GTK_BOX(int_hbox), interval_value, TRUE, TRUE, 0);
    gtk_widget_set_usize(interval_value, 100, -2);

    int_secs = gtk_label_new(_("seconds"));
    gtk_widget_ref(int_secs);
    gtk_object_set_data_full(GTK_OBJECT(properties), "int_secs", int_secs,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(int_secs);
    gtk_box_pack_start(GTK_BOX(int_hbox), int_secs, FALSE, FALSE, 0);

    messages_check = gtk_check_button_new_with_label(_("Show number of messages"));
    gtk_widget_ref(messages_check);
    gtk_object_set_data_full(GTK_OBJECT(properties), "messages_check", 
                             messages_check,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(messages_check);
    gtk_box_pack_start(GTK_BOX(props_vbo), messages_check, FALSE, FALSE, 0);

    time_check = gtk_check_button_new_with_label(_("Show timestamp"));
    gtk_widget_ref(time_check);
    gtk_object_set_data_full(GTK_OBJECT(properties), "time_check", time_check,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(time_check);
    gtk_box_pack_start(GTK_BOX(props_vbo), time_check, FALSE, FALSE, 0);

    user_check = gtk_check_button_new_with_label(_("Show username"));
    gtk_widget_ref(user_check);
    gtk_object_set_data_full(GTK_OBJECT(properties), "user_check", user_check,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(user_check);
    gtk_box_pack_start(GTK_BOX(props_vbo), user_check, FALSE, FALSE, 0);

    button_hbox = gtk_hbox_new(TRUE, 0);
    gtk_widget_ref(button_hbox);
    gtk_object_set_data_full(GTK_OBJECT(properties), "button_hbox", 
                             button_hbox,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(button_hbox);
    gtk_box_pack_start(GTK_BOX(props_vbo), button_hbox, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(button_hbox), 5);

    ok_button = gtk_button_new_with_label (_("OK"));
    gtk_widget_ref(ok_button);
    gtk_object_set_data_full(GTK_OBJECT(properties), "ok_button", ok_button,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(ok_button);
    gtk_box_pack_start(GTK_BOX(button_hbox), ok_button, FALSE, FALSE, 0);

    apply_button = gtk_button_new_with_label (_("Apply"));
    gtk_widget_ref(apply_button);
    gtk_object_set_data_full(GTK_OBJECT(properties), "apply_button", 
                             apply_button,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(apply_button);
    gtk_box_pack_start(GTK_BOX(button_hbox), apply_button, FALSE, FALSE, 0);

    cancel_button = gtk_button_new_with_label (_("Cancel"));
    gtk_widget_ref(cancel_button);
    gtk_object_set_data_full(GTK_OBJECT(properties), "cancel_button", 
                             cancel_button,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(cancel_button);
    gtk_box_pack_start(GTK_BOX(button_hbox), cancel_button, FALSE, FALSE, 0);

    gtk_signal_connect(GTK_OBJECT(properties), "delete_event",
                       GTK_SIGNAL_FUNC(dismiss_props),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(mail_rb), "toggled",
                       GTK_SIGNAL_FUNC(mail_toggled),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(custom_rb), "toggled",
                       GTK_SIGNAL_FUNC(custom_toggled),
                       NULL);
    gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
                       GTK_SIGNAL_FUNC(ok_clicked),
                       NULL);
    gtk_signal_connect (GTK_OBJECT (apply_button), "clicked",
                       GTK_SIGNAL_FUNC (apply_clicked),
                       NULL);
    gtk_signal_connect (GTK_OBJECT (cancel_button), "clicked",
                       GTK_SIGNAL_FUNC (cancel_clicked),
                       NULL);

    set_props_state();

    return(properties);
}


/*ARGSUSED*/
static void
custom_toggled(GtkToggleButton *tb, gpointer user_data)
{
    set_mail_custom_widgets(!gtk_toggle_button_get_active(tb));
}


void
destroy_image(enum icon_type itype)
{
    if (images[(int) itype] != 0) {
        gdk_pixmap_unref(images[(int) itype]);
        images[(int) itype] = 0;
    }
}


/*ARGSUSED*/
static gboolean
dismiss_faces(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    exit(1);

    return(FALSE);
}


/*ARGSUSED*/
static gboolean 
dismiss_props(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    props_window = NULL;

    return(FALSE);
}


void
display_error(char *message)
{
    if (error_displayed == FALSE) {
        gtk_label_set_text(GTK_LABEL(error_label), message);
        gtk_widget_show(error_dialog);
        error_displayed = TRUE;
    } else {
        fprintf(stderr, "%s", message);
    }
}


static void
display_props_sheet()
{
    if (props_window == NULL) {
        props_window = create_properties();
    }
    gtk_window_set_position(GTK_WINDOW(props_window), GTK_WIN_POS_CENTER);
    gtk_widget_show(props_window);
}


void
display_warning(char *message)
{
    if (warning_displayed == FALSE) {
        gtk_label_set_text(GTK_LABEL(warning_label), message);
        gtk_widget_show(warning_dialog);
        warning_displayed = TRUE; 
    } else { 
        fprintf(stderr, "%s", message); 
    }
}


static void
do_error_dialog_ok(GtkButton *button, gpointer user_data)
{
    gtk_widget_hide(error_dialog);
    error_displayed = FALSE;
    if (have_login_info == FALSE && login_timerid == 0) {
        login_timerid = gtk_timeout_add(500, display_login_popup, NULL);
    }
}


static void
do_warning_dialog_ok(GtkButton *button, gpointer user_data)
{
    gtk_widget_hide(warning_dialog);
    warning_displayed = FALSE;
}


#ifdef IMAP
static void
do_imap_login(GtkButton *button, gpointer user_data)
{
    imap_machine = gtk_entry_get_text(GTK_ENTRY(imap_server));
    imap_user    = gtk_entry_get_text(GTK_ENTRY(imap_login));
    imap_pwd     = gtk_entry_get_text(GTK_ENTRY(imap_password));

    gtk_widget_hide(imap_popup);

    have_login_info = TRUE;
    do_check();                  /* Generate initial faces display. */

    if (have_login_info == TRUE) {
        timerid = gtk_timeout_add(period * 1000, update_timer, NULL);
    }
}


static void
do_imap_cancel(GtkButton *button, gpointer user_data)
{
    FPRINTF(stderr, "IMAP login cancelled. Terminating.\n");
    exit(1);
}
#endif /*IMAP*/


static void
do_pause()
{
    static struct timeval btime = { 0, 250000 };   /* Beep timer. */

    gdk_flush();
    SELECT(0, (fd_set *) 0, (fd_set *) 0, (fd_set *) 0, &btime);
}


#ifdef POP
static void
do_pop_login(GtkButton *button, gpointer user_data)
{
    pop_machine = gtk_entry_get_text(GTK_ENTRY(pop_server));
    pop_user    = gtk_entry_get_text(GTK_ENTRY(pop_login));
    pop_pwd     = gtk_entry_get_text(GTK_ENTRY(pop_password));

    gtk_widget_hide(pop_popup);

    have_login_info = TRUE;
    do_check();                  /* Generate initial faces display. */

    if (have_login_info == TRUE) {
        timerid = gtk_timeout_add(period * 1000, update_timer, NULL);
    }
}

static void
do_pop_cancel(GtkButton *button, gpointer user_data)
{
    FPRINTF(stderr, "POP login cancelled. Terminating.\n");
    exit(1);
}
#endif /*POP*/


void
draw_screen(enum disp_type wdtype)
{
    gdk_window_copy_area(canvas_win, gc, 0, 0, pr[(int) wdtype], 0, 0,
                         (unsigned int) width, (unsigned int) height);

    gdk_window_set_icon(window->window, NULL, pr[(int) DISP_ICON], iconbm);
}


void
drop_back()
{
    if (lowerwindow) {
        gdk_window_lower(window->window);
    }
}


/* Get faces resource from merged databases. */

char *
get_resource(enum res_type rtype)
{
    char cstr[MAXLINE], nstr[MAXLINE], str[MAXLINE];
    char *str_type[20];
    XrmValue value;

    STRCPY(str, resources[(int) rtype]);
    SPRINTF(nstr, "faces.%s", str);
    if (islower((int) str[0])) {
        str[0] = toupper(str[0]);
    }

    SPRINTF(cstr, "Faces.%s", str);
    if (XrmGetResource(faces_DB, nstr, cstr, str_type, &value) == 0) {
        return((char *) NULL);
    } else {
        return(value.addr);
    }
}


static unsigned short
get_hex(FILE *fp)
{
    int c;
    unsigned short rval = 0;
 
    while ((c = getc(fp)) != EOF) {
        if (c != '0') {
            continue;
        }
        c = getc(fp);
        if (c != 'x') {
            continue;
        }
        while ((c = getc(fp)) != EOF) {
            if (c >= '0' && c <= '9') {
                rval = (rval << 4) + (c - '0');
            } else if (c >= 'a' && c <= 'f') {
                rval = (rval << 4) + (c - 'a' + 10);
            } else if (c >= 'A' && c <= 'F') {
                rval = (rval << 4) + (c - 'A' + 10);
            } else {
                break;
            }
        }
        break;
    }
 
    return(rval);
}


int
get_xbm_icon(char *name, enum icon_type itype)    /* Load XBM icon file. */
{
    FILE *fin;
    int hgt, i, j, wid;
    char c, *cptr;
    unsigned char vbuf[512];
    unsigned int tmp;
    unsigned short buf[256];

    if ((fin = fopen(name, "r")) == NULL) {
        return(-1);
    }
    if (fgets(nextline, MAXLINE, fin) == NULL) {
        goto err_end;
    }
    if (sscanf(nextline, "#define %s %d", vbuf, &wid) != 2) {
        goto err_end;
    }
    if ((cptr = strrchr((char *) vbuf, '_')) == NULL || 
        strcmp(cptr, "_width")) {
        goto err_end;
    }
    wid = (wid + 7) / 8;

    if (fgets(nextline, MAXLINE, fin) == NULL) {
        goto err_end;
    }
    if (sscanf(nextline, "#define %s %d", vbuf, &hgt) != 2) {
        goto err_end;
    }
    if ((cptr = strrchr((char *) vbuf, '_')) == NULL || 
        strcmp(cptr, "_height")) {
        goto err_end;
    }
    if (hgt > iconheight) {
        hgt = iconheight;
    }

    while (nextline[0] == '#') {
        if (fgets(nextline, MAXLINE, fin) == NULL) {
            goto err_end;
        }
        if (sscanf(nextline, "static %s %*[^{]%c", vbuf, &c) != 2) {
            goto err_end;
        }
        if (strcmp((char *) vbuf, "char") == 0) {
            for (i = 0; i < sizeof(vbuf); vbuf[i++] = 0) {
                continue;
            }
            for (i = 0; i < hgt; i++) {
                for (j = 0; j < wid; j++) {
                    tmp = get_hex(fin) & 0xFF;
                    if (j < (iconwidth / 8)) {
                        vbuf[i*8+j] = (unsigned char) tmp;
                    }
                }
            }
            for (i = 0; i < 256; i++) {
                buf[i] = (revtable[vbuf[(i*2)+1]] & 0xFF) +
                         ((revtable[vbuf[i*2]] & 0xFF) << 8);
            }
            FCLOSE(fin);

            load_xbm_icon(itype, buf);
            return(0);
        } else if (strcmp((char *) vbuf, "short") == 0) {
            wid = (wid + 1) / 2;
            for (i = 0; i < 256; buf[i++] = 0) {
                continue;
            }
            for (i = 0; i < hgt; i++) {
                for (j = 0; j < wid; j++) {
                    tmp = get_hex(fin);
                    if (j < (iconwidth / 16)) {
                        buf[i*4+j] = (revtable[tmp >> 8] & 0xFF) +
                                     ((revtable[tmp & 0xFF] & 0xFF) << 8);
                    }
                }
            }
            FCLOSE(fin);

            load_xbm_icon(itype, buf);
            return(0);
        }
    }

err_end:

    FCLOSE(fin);

    return(-1);
}


int
get_xpm_icon(char *name, enum icon_type itype)   /* Load XPM icon file. */
{
    GdkBitmap *mask;
    GdkPixmap *pixmap;

    if ((pixmap = gdk_pixmap_create_from_xpm(GDK_ROOT_PARENT(), &mask,
                                             NULL, name)) == NULL) {
        return(-1);
    }

    images[(int) itype] = pixmap;

    return(0);
}


static void
init_font()
{
    if (fontname != NULL) {
        sfont = gdk_font_load(fontname);
    }
    if (sfont != NULL) {
        return;
    }

    if ((sfont = gdk_font_load(FONT)) == NULL) {
        if ((sfont = gdk_font_load(DEF_FONT)) == NULL) {
            perror("couldn't get the default font.");
            exit(1);
        }
    }
}


/*ARGSUSED*/
void
init_graphics(int *argc, char *argv[])
{
#ifdef ENABLE_NLS
    bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
    textdomain(PACKAGE);
#endif

    gtk_set_locale();
    gtk_init(argc, &argv);
    dpy = GDK_DISPLAY();

    faces_DB = NULL;
}


/*ARGSUSED*/
int
init_ws_type(int argc, char *argv[])
{
    int dummy, i, j;
    unsigned short buf[256];
    GdkColor selcolor;

    i = XParseGeometry(icon_geometry, &ix, &iy,
                       (unsigned int *) &j, (unsigned int *) &j);
    if (i & (XValue | YValue)) {
        if (i & XNegative) {
            ix = gdk_screen_width() + ix - imagewidth;
        }   
        if (i & YNegative) {
            iy = gdk_screen_height() + iy - imageheight;
        }   
    }

    cmap = gdk_colormap_get_system();
    gdk_window_get_geometry(GDK_ROOT_PARENT(),
                            &dummy, &dummy, &dummy, &dummy, &depth);
    if (fgcolor == NULL || depth <= 2) {
        (void) gdk_color_black(cmap, &foregnd);
    } else {
        gdk_color_parse(fgcolor, &selcolor);
        gdk_colormap_alloc_color(cmap, &selcolor, TRUE, TRUE);
        foregnd = selcolor;
    }        
    if (bgcolor == NULL || depth <= 2) {
        (void) gdk_color_white(cmap, &backgnd);
    } else { 
        gdk_color_parse(bgcolor, &selcolor);
        gdk_colormap_alloc_color(cmap, &selcolor, TRUE, TRUE);
        backgnd = selcolor;
    }        

    init_font();
    gc_mask = GDK_GC_FONT | GDK_GC_FOREGROUND |
              GDK_GC_BACKGROUND | GDK_GC_EXPOSURES;
    gc_val.font = sfont;
    gc_val.foreground = foregnd;
    gc_val.background = backgnd;
    gc_val.graphics_exposures = False;
    gc = gdk_gc_new_with_values(GDK_ROOT_PARENT(), &gc_val, gc_mask);

    xpmgc = gdk_gc_new_with_values(GDK_ROOT_PARENT(), &gc_val, gc_mask);
    gdk_gc_set_clip_origin(xpmgc, 0, 0);

    tilegc = gdk_gc_new_with_values(GDK_ROOT_PARENT(), &gc_val, gc_mask);
    if (depth == 1) {
        gdk_gc_set_fill(tilegc, GDK_OPAQUE_STIPPLED);
    } else {
        gdk_gc_set_fill(tilegc, GDK_TILED);
    }

    for (i = 0; i < 16; i++) {     /* Load default gray background. */
        for (j = 0; j < 4; j++) {
            buf[i*16      + j] = 0x7777;
        }
        for (j = 0; j < 4; j++) {
            buf[i*16 +  4 + j] = 0xDDDD;
        }
        for (j = 0; j < 4; j++) {
            buf[i*16 +  8 + j] = 0xBBBB;
        }
        for (j = 0; j < 4; j++) {
            buf[i*16 + 12 + j] = 0xEEEE;
        }
    }
 
    if (bgicon != NULL) {
        (void) get_xbm_icon(bgicon, BACKIMAGE);
    } else {
        load_xbm_icon(BACKIMAGE, buf);
    }

    STRCPY(fname[0], "face.xpm");
    STRCPY(fname[1], "face.xbm");
    maxtypes = 2;
    pr[(int) DISP_NAME]     = pr[(int) DISP_OTHER]     = 0;
    old_pr[(int) DISP_NAME] = old_pr[(int) DISP_OTHER] = 0;

    return(0);
}


void
load_data(enum icon_type itype, unsigned char *cbuf)
{
    images[(int) itype] = gdk_pixmap_create_from_data(canvas_win,
                                    (char *) cbuf, 
                                    iconwidth, iconheight, depth,
                                    &foregnd, &backgnd);
}


/*ARGSUSED*/
void
load_xbm_icon(enum icon_type itype, unsigned short sbuf[256])
{
    char cbuf[512];
    int i;

    for (i = 0; i < 256; i++) {
        cbuf[i*2+0] = revtable[(sbuf[i] >> 8) & 0xFF];
        cbuf[i*2+1] = revtable[sbuf[i] & 0xFF];
    }

    images[(int) itype] = gdk_pixmap_create_from_data(canvas_win, cbuf, 
                                           iconwidth, iconheight, depth,
                                           &foregnd, &backgnd);
}


/* Load the faces resources database from FACESDEFAULTS environment variable 
 * (if set), or from $HOME/.facesdefaults.
 */

void
load_resources()
{
    char name[MAXPATHLEN], *ptr;

    XrmInitialize();
    if (getenv("HOME") == NULL) {
        faces_DB = NULL;
        return;
    } else if ((ptr = getenv("FACESDEFAULTS")) == NULL) {
        SPRINTF(name, "%s/.facesdefaults", getenv("HOME"));
        ptr = name;
    }
    faces_DB = XrmGetFileDatabase(ptr);
}


static void
set_mail_custom_widgets(gboolean is_mail)
{
    if (is_mail) {
        gtk_label_set_text(GTK_LABEL(name_label),  _("Folder name:"));  
        gtk_widget_show(type_hbox);
    } else {
        gtk_label_set_text(GTK_LABEL(name_label),  _("Custom command:"));
        gtk_widget_hide(type_hbox);
    }
}


/*ARGSUSED*/
static void
mail_toggled(GtkToggleButton *tb, gpointer user_data)
{
    set_mail_custom_widgets(gtk_toggle_button_get_active(tb));
}


void
make_area(enum disp_type dtype, int width, int height)
{
    pr[(int) dtype] = gdk_pixmap_new(GDK_ROOT_PARENT(), width, height, depth);
    gdk_gc_set_function(gc, GDK_CLEAR);
    gdk_window_copy_area(pr[(int) dtype], gc, 
                         column*imagewidth, row*imagewidth,
                         pr[(int) dtype], column*imagewidth, row*imagewidth, 
                         imagewidth, imageheight);
    gdk_gc_set_function(gc, GDK_COPY);
}


/*ARGSUSED*/
void
make_frame(int argc, char *argv[])
{
    unsigned int h, w;                /* Window dimensions. */
    int flags;
    int x, y;                         /* Window position. */
    GdkBitmap *mask;
    GdkPixmap *pixmap;

    w = width = maxcols * imagewidth;
    h = height = imageheight;

    size.flags = PMinSize | PMaxSize | PPosition | PSize;
    size.x = 0;
    size.y = 0;
    size.min_width = imagewidth;
    size.min_height = imageheight;
    size.max_width = size.width = width;
    if (mtype == MONNEW) {
        size.max_height = imageheight;
    } else {
        size.max_height = 900;
    }
    size.height = height;

    if (geometry != NULL) {
        flags = XParseGeometry(geometry, &x, &y, &w, &h);
        if (XValue & flags) {
            if (XNegative & flags) {
                x = gdk_screen_width() + x - size.width;
            }
            size.flags |= USPosition;
            size.x = x;
        }
        if (YValue & flags) {
            if (YNegative & flags) {
                y = gdk_screen_height() + y - size.height;
            }
            size.flags |= USPosition;
            size.y = y;
        }
    }

    ix = size.x + size.width - imagewidth;
    iy = size.y;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data(GTK_OBJECT(window), "window", window);
    gtk_widget_set_usize(window, size.width, size.height);
    gtk_window_set_title(GTK_WINDOW(window), _("faces"));
    gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);

    canvas = gtk_drawing_area_new();
    gtk_widget_ref(canvas);
    gtk_object_set_data_full(GTK_OBJECT(window), "canvas", canvas,
                             (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(canvas);
    gtk_container_add(GTK_CONTAINER(window), canvas);

    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
                       GTK_SIGNAL_FUNC(key_press), NULL);
    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
                       GTK_SIGNAL_FUNC(dismiss_faces), NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "expose_event",
                       GTK_SIGNAL_FUNC(process_expose), NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "button_press_event",
                       GTK_SIGNAL_FUNC(button_press), NULL);
    gtk_signal_connect(GTK_OBJECT(canvas), "configure_event",
                       GTK_SIGNAL_FUNC(canvas_resize), NULL);

    gtk_widget_realize(canvas);
    gtk_widget_realize(window);
    canvas_win = canvas->window;

    iconbm = gdk_bitmap_create_from_data(window->window,
                                         (const gchar *) nomail_bits,
                                         nomail_width, nomail_height);

    pixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, 
                                          gtk_widget_get_default_colormap(),
                                          &mask, NULL, 
                                          (gchar **) dialog_error);
    errorpm = gtk_pixmap_new(pixmap, mask);

    pixmap = gdk_pixmap_colormap_create_from_xpm_d(NULL, 
                                          gtk_widget_get_default_colormap(),
                                          &mask, NULL, 
                                          (gchar **) dialog_warning);
    warningpm = gtk_pixmap_new(pixmap, mask);

    /* set the icon for the main window. */
    gdk_window_set_icon(window->window, NULL, iconbm, iconbm);
    gdk_window_set_icon_name(window->window, "faces");

    protocol_atom = gdk_atom_intern("WM_PROTOCOLS", FALSE);
    kill_atom     = gdk_atom_intern("WM_DELETE_WINDOW", FALSE);

    error_dialog   = create_dialog(ERROR,   "Error",   &error_label);
    warning_dialog = create_dialog(WARNING, "Warning", &warning_label);

    adjust_image(DISP_ICON, NOMAIL,  0, 0);

	if (!is_imap && !is_pop) {
		timerid = gtk_timeout_add(period * 1000, update_timer, NULL);
	}
}
 

static gboolean
button_press(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
    if (event->button == 3) {
        display_props_sheet();
    } else if (button1clear && event->button == 1) {
        do_key(KEY_CLEAR);
    } else if ((button1clear && event->button == 2) ||
               (!button1clear && event->button == 1)) {
        wdtype = (wdtype == DISP_NAME) ? DISP_OTHER : DISP_NAME;
        draw_screen(wdtype);
    }

    return(FALSE);
}


static gboolean
key_press(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
    do_key(event->keyval);

    return(FALSE);
}


static gboolean 
process_expose(GtkWidget *widget, GdkEventExpose *event, gpointer user_data)
{
    draw_screen(wdtype);

    if ((is_imap || is_pop) && have_login_info == FALSE && login_timerid == 0) {
        login_timerid = gtk_timeout_add(500, display_login_popup, NULL);
    }

    return(FALSE);
}


/*ARGSUSED*/
static gint
display_login_popup(gpointer data)
{
    if (login_timerid != 0) {
#ifdef IMAP
        if (is_imap) {
            imap_popup = create_imap_popup();
            gtk_widget_show(imap_popup);
            gtk_timeout_remove(login_timerid);
            login_timerid = 0;
        }
#endif /*IMAP*/

#ifdef POP
        if (is_pop) {
            pop_popup = create_pop_popup();
            gtk_widget_show(pop_popup);
            gtk_timeout_remove(login_timerid);
            login_timerid = 0;
        }
#endif /*POP*/
    }

    return(TRUE);
}


/*ARGSUSED*/
static gint
update_timer(gpointer data)
{
    if (timerid != 0) {
        do_check();     /* Check the mail queue again. */
    }

    return(TRUE);
}


static void
do_props_button_clicked(enum pbutton_type pbtype)
{
    switch (pbtype) {
        case P_OK:
        case P_APPLY:
            if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mail_rb))) {
                if (normal_rb) {
                    ftype = F_NORMAL;
                } else if (imap_rb) {
                    ftype = F_IMAP;
                } else if (pop_rb) {
                    ftype = F_POP;
                }

                spoolfile = strdup(gtk_entry_get_text(GTK_ENTRY(name)));
            } else {
                userprog = strdup(gtk_entry_get_text(GTK_ENTRY(name)));
            }    
 
            period = gtk_spin_button_get_value_as_int(
                                             GTK_SPIN_BUTTON(interval_value));
            dontshowno = gtk_toggle_button_get_active(
                                           GTK_TOGGLE_BUTTON(messages_check));
            dontshowtime = gtk_toggle_button_get_active(
                                               GTK_TOGGLE_BUTTON(time_check));
            dontshowuser = gtk_toggle_button_get_active(
                                               GTK_TOGGLE_BUTTON(user_check));

            write_resources();
            if (pbtype == P_OK) {
                gtk_widget_hide(props_window);
            }
            break;

        case P_CANCEL:
            gtk_widget_hide(props_window);
            break;
    }
}


static void
ok_clicked(GtkButton *button, gpointer user_data)
{
    do_props_button_clicked(P_OK);
}


void
put_resource(enum res_type rtype, char *rval)
{
    char app[MAXLINE], resource[MAXLINE];

    STRCPY(app, progname);
    if (isupper((int) app[0])) {
        app[0] = tolower(app[0]);
    }
    SPRINTF(resource, "%s.%s", app, resources[(int) rtype]);
    XrmPutStringResource(&faces_DB, resource, (rval) ? rval : "");
}


void
repl_image(enum disp_type dtype, enum image_type dest, int width, int height)
{
    GdkPixmap *dpm;
 
    if (dest == CUROFF) {
        dpm = pr[(int) dtype];
    } else {
        dpm = old_pr[(int) dtype];
    }   
    if (depth == 1) {
        gdk_gc_set_stipple(tilegc, images[(int) BACKIMAGE]);
    } else {
        gdk_gc_set_tile(tilegc, images[(int) BACKIMAGE]);
    }   
    gdk_draw_rectangle(dpm, tilegc, TRUE, 0, 0, width, height);
}


int
save_resources(void)
{
    char *home, *filename;
    struct stat statbuf;

    if ((home = getenv("HOME")) == NULL) {
        return(1);
    }

    if ((filename = getenv("FACESDEFAULTS")) == NULL) {
        filename = (char *) calloc(1, strlen(home) + 18);
        SPRINTF(filename, "%s/.facesdefaults", home);
    }

/* If file exists but user does not have access. */

    if (stat(filename, &statbuf) != -1 && access(filename, W_OK) != 0) {
        FREE(filename);
        return(1);
    }

/* If file does not exist this call will create it. */

    XrmPutFileDatabase(faces_DB, filename);
    FREE(filename);
    return(0);
}


static void
set_props_state()
{
    gboolean is_mail = (mtype != MONPROG);

    if (is_mail) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mail_rb), TRUE);
        if (is_imap) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(imap_rb), TRUE);
        } else if (is_pop) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(pop_rb), TRUE);
        } else {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(normal_rb), TRUE);
        }
        if (spoolfile != NULL) {
            gtk_entry_set_text(GTK_ENTRY(name), spoolfile);
        }
    } else {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(custom_rb), TRUE);
        gtk_entry_set_text(GTK_ENTRY(name), userprog);
    }

    set_mail_custom_widgets(is_mail);

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(interval_value), (gfloat) period);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(messages_check), !dontshowno);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(time_check), !dontshowtime);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(user_check), !dontshowuser);
}


/* Show the latest set of mail icon faces. */

void
show_display(int beep_and_raise)
{
    static int first_time = 1;

    if (invert) {     /* Invert the memory pixrects before displaying. */
        gdk_gc_set_function(gc, GDK_COPY_INVERT);
        gdk_window_copy_area(pr[(int) DISP_NAME], gc, 0, 0,
                             pr[(int) DISP_NAME], 0, 0,
                             (unsigned int) width, (unsigned int) height);
        gdk_window_copy_area(pr[(int) DISP_OTHER], gc, 0, 0,
                             pr[(int) DISP_OTHER], 0, 0,
                             (unsigned int)  width, (unsigned int) height);
    }
    if (first_time) {
        gdk_window_set_events(canvas_win, FRAME_MASK);
        gdk_window_show(canvas_win);
        first_time = 0;
    }   

    if (beep_and_raise && raisewindow && newmail && !mboxshrunk) {
        gdk_window_raise(window->window);
    }   
 
    gtk_drawing_area_size(GTK_DRAWING_AREA(canvas), width, height);
    gtk_widget_set_usize(window, width, height);
    draw_screen(wdtype);
 
    if (beep_and_raise && newmail && !mboxshrunk) {
        beep_flash(beeps, flashes);
    }   
    if (old_pr[(int) DISP_NAME]) {
        gdk_pixmap_unref(old_pr[(int) DISP_NAME]);
    }   
    if (old_pr[(int) DISP_OTHER]) {
        gdk_pixmap_unref(old_pr[(int) DISP_OTHER]);
    }   
    old_pr[(int) DISP_NAME]  = 0;
    old_pr[(int) DISP_OTHER] = 0;
    gdk_flush();
}
 

void
start_tool()
{
    gtk_widget_show(window);
    gtk_main();
}


void
text(enum disp_type dtype, enum just_type jtype, char *str)
{
    int len;
    int c, r;           /* Column and row position for this face. */
    int x = 0, y = 0;   /* Position of start of this text string. */
    int ascent, descent, dummy, textheight, textwidth;
 
    c = column;
    r = row;
    switch (dtype) {
        case DISP_ALL : 
            text(DISP_ICON,  jtype, str);
            /*FALLTHROUGH*/
        case DISP_BOTH : 
            text(DISP_NAME,  jtype, str);
            text(DISP_OTHER, jtype, str);
            return;
        case DISP_ICON : 
            c = r = 0;
            break;
        default :
            break;
    }
 
    if ((len = strlen(str)) > 20) {
        len = 20;                       /* Character length of text. */
    }

    do {
        gdk_text_extents(sfont, str, len, &dummy, &dummy, 
                         &textwidth, &ascent, &descent);

        textheight = ascent + descent;
        if (textwidth < imagewidth) {
            break;
        }
        len--;
    } while (len > 5);

    switch (jtype) {
        case LEFT  : x =  c      * imagewidth + 2;
                     y = (r + 1) * imageheight - textheight - 5;
                     break;
        case RIGHT : x = (c + 1) * imagewidth - textwidth - 2;
                     y = (r + 1) * imageheight - textheight - 5;
    }

    gdk_gc_set_function(gc, GDK_COPY_INVERT);

    gdk_draw_rectangle(pr[(int) dtype], gc, TRUE, 
                       x, y, textwidth+2, textheight + 5);
    gdk_gc_set_function(gc, GDK_COPY);

    y = (r + 1) * imageheight - descent - 2;

    /* Make sure we at least draw something... */
    gdk_draw_text(pr[(int) dtype], sfont, gc, x+1, y, str, len);
}
