
/*  $Header: /cvs/faces/faces/faces/mon.c,v 1.31 2002/03/06 11:27:08 devface Exp $
 *
 *  Monitoring routines used by the faces program.
 * 
 *  Copyright (C) 1990-2002  Rich Burridge, Sun Microsystems Inc.
 *  All rights reserved.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 * 
 *  No responsibility is taken for any errors on inaccuracies inherent
 *  either to the comments or the code of this program, but if reported
 *  to me, then an attempt will be made to fix them.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <utime.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "faces.h"
#include "extern.h"

#ifdef MDEBUG
#define  M_DEBUG(stmt)  stmt
#else
#define  M_DEBUG(stmt)
#endif /*MDEBUG*/

#ifdef POP
#include "pop.h"
/* POP variables */
long* mailItemArray;	/* malloc'ed into an array of longs */
int mIAIndex;
int pop_processed;		/* flags that From has been processed */
static long pop_last_messages = 0;	/* Used to remember mailbox size */
static long pop_last_boxsize = 0;	/*   so we know when new mail as arrived */
time_t pop_this_date = 0;			/* date of the current processed mail*/
#endif /* POPMAIL */
#ifdef IMAP
#include "imap.h"
/* IMAP variables */
int imap_processed;		/* flags that From has been processed */
static long imap_last_messages = 0;	/* Used to remember mailbox size */
static long imap_last_boxsize = 0;	/* so we know when new mail as arrived */
time_t imap_this_date = 0;			/* date of the current processed mail*/
#endif /* IMAP */

int m_eoh;		/* flag to remember outside of the header */
static long* idmem = NULL;
static int idmem_last = 0;
static int idmem_size = 0;

/* function protoypes */
extern int parse822from(char *, char **, char**);
extern int uncompface(char *);

void do_check();
void do_key(int);
static void do_mail(enum mon_type);

static void do_mailspool(enum mon_type);
int LWSP(int);
static char * fgetheader(int, FILE *);
static int casecmp(int, int);
int is822header(char *, char *);

#ifdef POP
static void do_mailPOP(enum mon_type);
static int collect_list(char*, void*);
static int pop_header(char*, void*);
void do_pop_process(enum mon_type);
#endif /* POP */

#ifdef IMAP
static void do_mailIMAP(enum mon_type);
int collect_list(char*, void*);
int pop_header(char*, void*);
int imap_collect_list(int, char*, void*);
int imap_header(int, char*, void*);
void do_imap_process(enum mon_type);
#endif	/* IMAP */

static void do_prog(enum mon_type);
static void get_userrec(char *, enum disp_type);
static void add_face(enum disp_type, enum icon_type, char *);
static void adjust(enum mon_type);

#ifdef AUDIO_SUPPORT
void check_audio(char *);
void play_sound(char *);
#endif /*AUDIO_SUPPORT*/

static void make_pixrect(int);
static void make_display(enum mon_type);
static void process_face();
static void process_from();
static void process_info(enum mon_type);
static void remember(long);
static int haveseen(long);

#define SKIPWS(c) while (*(c) && isspace((unsigned char) *(c))) c++;
#define MAX(a,b) ((a)>(b)?(a):(b))

/* Select the type of monitoring we're doing */
void
do_check()
{
    switch (mtype) {
        case MONALL     : do_mail(MONALL);  /* Monitor all of the mail file. */
                          break;
        case MONNEW     : do_mail(MONNEW);  /* Monitor new mail only. */
                          break;
        case MONPROG    : do_prog(mtype);        /* Run user supplied program. */
                          break;
    }
    firsttime = 0;
}


/* Handle various keyboard options. */
void
do_key(int val)
{
    switch (val) {
        case KEY_CLEAR   : if (mtype == MONNEW) {
                               repl_image(DISP_NAME,  CUROFF, width, height);
                               repl_image(DISP_OTHER, CUROFF, width, height);
                               add_face(DISP_ICON, NOMAIL, "");
                               toclear = 1;
                           }
                           do_check();
                           break;

        case KEY_REFRESH : draw_screen(wdtype);
                           break;

        case KEY_QUIT    : exit(0);
    }
}

/* Supposed to be checking mail.
 * Select the type of mail to scan */
static void
do_mail(enum mon_type mtypel)
{
#ifdef POP
	if (is_pop)
		do_mailPOP(mtypel);
#endif
#ifdef IMAP
	if (is_imap)
		do_mailIMAP(mtypel);
#endif
	if (!is_pop && !is_imap)
		do_mailspool(mtypel);
}

/* ==MAIL SPOOL FILE============================================ */

/* Monitor a mail spool file for new or all mail. */
/* 
 * After first checking to see if the spool file changed since we
 * last looked at it, this routine reads through the spool file
 * looking for "From ", "From:" and "X-Face:: lines.
 * For each of the found header lines, this calls either 'process_from()'
 * or 'process_face()'.  These routines populate the global variables
 * 'face_user', 'face_host' and 'pict_buf'. Between the calls to these
 * routines, this calles 'process_info()' to map the user/host to a
 * picon and build the output pixmap.  */
static void
do_mailspool(enum mon_type mtypel)
{        
    int eoh;                 /* Set true, if end of mail header detected. */
    FILE *fp;                /* File descriptor for users mail spool file. */
    struct utimbuf ubuf;     /* For resetting the access time on spoolfile. */
    static int first = 1;    /* First time through. */

    column = row = 0;        /* Start in top left corner of pixrect. */
    mboxshrunk = 0;
    newmail = 0;             /* Assume no new mail. */
    noicons = 0;

    if (spoolfile == NULL) {
        return;
    }

	/* fake liked logged on for error dialogs and such */
	have_login_info = TRUE;

    if (mtypel == MONNEW) {
        make_pixrect(maxcols);
    }
    if (stat(spoolfile, &buf) == -1) {
		M_DEBUG(FPRINTF(stderr, "do_mailspool: no mail\n"));
        lastsize = 0;
        add_face(DISP_ICON, NOMAIL, "");
        if (mtypel == MONNEW) {
            show_display(TRUE);  /* Show new mail. */
        } else {
            make_display(mtypel);   /* Output icons and tidyup chain of records. */
        }
        goto spoollogoutandreturn;
    }

/*  See if there is any work to do. If the file hasn't changed, then no need
 *  to do anything but exit. A size check is now done, because the Elm mailer
 *  resets the files' mtime.
 */

    if (!first && buf.st_size == lastsize) {
		M_DEBUG(FPRINTF(stderr, "do_mailspool: file size the same\n"));
        show_display(FALSE);
        goto spoollogoutandreturn;
    }

    first = 0;

    if (buf.st_size > lastsize ||
        (mhflag && (buf.st_size < lastsize) && (buf.st_size != 0))) {
        if (buf.st_size < lastsize) {
			M_DEBUG(FPRINTF(stderr, "do_mailspool: mailbox shrunk\n"));
            lastsize = 0;
            mboxshrunk = 1;
        }
        newmail = 1;                           /* New mail found. */
    }
    if (!buf.st_size) {
        add_face(DISP_ICON, NOMAIL, "");
    }

    CLEARMEM((char *) &ubuf, sizeof(ubuf)); /* Some systems have extra fields */
    ubuf.actime  = buf.st_atime;              /* Save for possible reset. */
    ubuf.modtime = buf.st_mtime;

    if (mtypel == MONNEW) {
        if (buf.st_size <= lastsize) {  /* Is the size of mail folder bigger? */
			M_DEBUG(FPRINTF(stderr, "do_mailspool: mailbox shrunk 2\n"));
            lastsize = buf.st_size;     /* No: save new size and return. */
            show_display(FALSE);
            goto spoollogoutandreturn;
        }
    }

    if ((fp = fopen(spoolfile, "r")) == NULL) {   /* Open spoolfile. */
		M_DEBUG(FPRINTF(stderr, "do_mailspool: could not open spool file\n"));
        if (mtypel == MONNEW) {
            show_display(FALSE); /* Show new mail. */
        } else {
            make_display(mtypel);      /* Output icons and tidyup chain of records. */
        }
        goto spoollogoutandreturn;
    }

    if (mtypel == MONNEW) {
        FSEEK(fp, lastsize, 0);
    }
    lastsize = buf.st_size;
    eoh = FALSE;
    while (fgetheader(eoh, fp) != NULL) {
        if (EQUAL(line, "From ")) {
            if (froms_found) {
                process_info(mtypel);  /* Process previous mail. */
            }
            eoh = FALSE;
            process_from();                   /* Save new from details. */
        } else if (*line == '\n' && eoh == FALSE) {
            eoh = TRUE;
        } else if (is822header(line, "From") && eoh == FALSE) {
            process_from();
        } else if (is822header(line, "X-Face") && eoh == FALSE) {
            process_face();
        }
    }
    FCLOSE(fp);

    if (!stat(spoolfile, &buf)) {  /* Reset access time if no change in size. */
        if (buf.st_size == lastsize) {
            UTIME(spoolfile, &ubuf);
        }
    }

    if (froms_found) {
        process_info(mtypel);          /* Process remaining mail item. */
    }
    if (mtypel == MONNEW) {
        show_display(TRUE);      /* Show new mail. */
    } else {
        make_display(mtypel);         /* Output icons and tidyup chain of records. */
    }
spoollogoutandreturn:
	return;
}

/* Return TRUE if passed character is white-space char (space or tab) */
int
LWSP(int c)
{
    return((c == ' ') || (c == '\t'));
}

/* Return either one line, or if eoh is FALSE, a folded RFC822 logical line. */
static char *
fgetheader(int eoh, FILE *fp)
{
    if (line == NULL) {
        line = Malloc(MAXLINE);
        maxline = MAXLINE;
    }
    if (eoh == TRUE) {
        /* just get a single line */
        return(fgets(line, maxline, fp));
    } else {
        /* get an entire RFC822 folded header */
        char *p;
        int c;

        /* get first line */
        if (fgets(line, maxline, fp) == NULL) {
            return(NULL);
        }

        /* check for EOH */
        if (line[0] == '\n') {
            return(line);
        }

        /* check for LWSP between field-name and initial : */
        if ((p = strchr(line, ':')) != NULL) {
            /* canonicalize */
            while (p > line && LWSP(p[-1])) {
                p[0] = p[-1];
                p[-1] = ':';
                p--;
            }
        }

        /* process continuation lines */
        for (;;) {
            /* get last char */
            p = line + strlen(line) - 1;
            if (*p != '\n') {
                /* this is a physically long line */
                p++;
                c = ' ';
            } else {
                c = getc(fp);
                if (c != EOF)
                    UNGETC(c,fp);
            }
            if (!LWSP(c)) {
                break;
            }

            /* this is a header continuation line */
            if ((p-line)+MAXLINE > maxline) {
                /* extend line buffer */
                maxline += MAXLINE;
                if ((line = realloc(line, maxline)) == NULL) {
                    FPRINTF(stderr, "Out of memory!\n");
                    exit(1);
                }
                p = line + strlen(line) - 1;
                if (*p != '\n') {
                    p++;
                }
            }
            p[0] = '\0';
            if ((p = fgets(p, MAXLINE, fp)) == NULL) {
                break;
            }
        }
        return(line);
    }
}


/* do a non-internationalized, case-less comparison of characters
 * Return 0 if two characters equal.  Non-zero otherwise.
 *
 * (non-internationlized is OK because only used to parse RFC822 headers) */
static int
casecmp(int c, int d)
{
    if (isupper(c)) {
        c = tolower(c);
    }
    if (isupper(d)) {
        d = tolower(d);
    }
    return(c - d);
}

/* Return non-zero if passed line properly begins with passed RFC822 header field.
 * Returns zero otherwise.
 * Checks properly for case and terminating colon (don't pass the latter)
 */
int
is822header(char *line, char *field)
{
    int n;

    n = strlen(field);
    while (n > 0) {
        if (casecmp(*line++, *field++) != 0) {
            return(0);
        }
        n--;
    }
    while (LWSP(*line)) {
        line++;
    }
    if (*line != ':') {
        return(0);
    }
     
    return(1);
}

#ifdef POP
/* == POP MAILBOX ============================================ */

static void
do_mailPOP(enum mon_type mtypel)
{
	int ret;
	POP_DATA* popdata;
	char buf[LONG_STRING];
	int i;
	int memarrsz, startind;
	long sizeThisTime, sizeLastTime;
	long boxsizeThisTime, boxsizeLastTime;

	/* start with new positions for picons */
    column = row = 0;        /* Start in top left corner of pixrect. */
    mboxshrunk = 0;
    newmail = 0;             /* Assume no new mail. */
    noicons = 0;

	/* Make sure the user has specified a server and login */
	if ((pop_machine == NULL) || (strlen(pop_machine)==0)) {
		M_DEBUG(FPRINTF(stderr, "do_mailPOP: pop_machine not set. Exitting\n"));
		return;
	}

	mailItemArray = NULL;

	/* if doing a new scan, start afresh */
    if (mtypel == MONNEW) {
        make_pixrect(maxcols);
    }

	/* Connect to the POP server */
	popdata = build_popdata(pop_machine, POP_PORT, pop_user, pop_pwd);
	ret = pop_open_connection(popdata);
	if (ret != POP_A_SUCCESS) {
		/* we need to tell the user something is wrong */
		/* TODO: error message passing and display */
		display_error("Could not connect or login to POP server");
		have_login_info = FALSE;
		goto logoutandbail;
	}
	have_login_info = TRUE;

	/* STAT the mailbox and see if it has changed */
	/* (STAT returns the number of messages and malbox size in octets) */
	ret = pop_get_statinfo(popdata, &sizeThisTime, &boxsizeThisTime);
	if (ret != POP_A_SUCCESS) {
		goto logoutandbail;
	}
	M_DEBUG(FPRINTF(stderr,"do_mailPOP: msgs=%ld, boxsize=%ld\n", 
				sizeThisTime, boxsizeThisTime));
	sizeLastTime = pop_last_messages;
	boxsizeLastTime = pop_last_boxsize;
	pop_last_messages = sizeThisTime;
	pop_last_boxsize = boxsizeThisTime;

	/* if the STAT of the mailbox has not changed, we don't need to rescan */
	if ((sizeThisTime == sizeLastTime) && (boxsizeThisTime == boxsizeLastTime)) {
		M_DEBUG(FPRINTF(stderr,"do_mailPOP: no changes, show_display(FALSE)\n"));
		show_display(FALSE);
		goto logoutandbail;
	}
	if (sizeThisTime > sizeLastTime) {
		M_DEBUG(FPRINTF(stderr,"do_mailPOP: new mail to add\n"));
		newmail = 1;
	}
	else {
		M_DEBUG(FPRINTF(stderr,"do_mailPOP: mailbox shrinking, start over\n"));
		mboxshrunk = 1;
	}

	/* if no mail in the mailbox, we can just put up NOMAIL */
	if (sizeThisTime == 0) {
		M_DEBUG(FPRINTF(stderr,"do_mailPOP: no mail, add_face(NOMAIL)\n"));
        add_face(DISP_ICON, NOMAIL, "");
        if (mtypel == MONNEW) {
			M_DEBUG(FPRINTF(stderr,"do_mailPOP: make_display(TRUE)\n"));
            show_display(TRUE);  /* Show new mail. */
        } else {
			M_DEBUG(FPRINTF(stderr,"do_mailPOP: make_display()\n"));
            make_display(mtypel);      /* Output icons and tidyup chain of records. */
        }
        goto logoutandbail;
	}

	/* Loop through the list of mail items and get their numbers */
	/* First, allocate an array to hold all of the message IDs */
	memarrsz = sizeThisTime*sizeof(long);
	mailItemArray = (long*)Malloc(memarrsz);
	mIAIndex = 0;
	CLEARMEM((char*)mailItemArray, memarrsz);
	ret = pop_fetch_data(popdata, "LIST", NULL, collect_list, NULL);
	if (ret != POP_A_SUCCESS) {
		/* TODO: for some reason can't list the mailbox.  WHat to tell the user? */
		goto logoutandbail;
	}

	/* Loop through the mail numbers and fetch their headers */
	/* if MONNEW, we only add new mail to the display */
	startind = mtypel == MONNEW ? sizeLastTime : 0;
	if (mtypel == MONNEW) {
		/* if MONNEW, only display the last 'maxcols' entries */
		startind = MAX(startind, mIAIndex-maxcols);
	}

	if (line)
		FREE(line);
	line = Malloc(MAXLINE);
	for (i=startind; i<mIAIndex; i++) {
		m_eoh = FALSE;
		pop_processed = FALSE;
		sprintf(buf, "TOP %ld 1", mailItemArray[i]);
		ret = pop_fetch_data(popdata, buf, NULL, pop_header, (void*)mtypel);
		/* two reasons for failure -- doesn't do TOP or message deleted
		 * A problem since missing message should be ignored but not doing
		 * TOP is a reason for total failure */
		if (ret != POP_A_SUCCESS) {
			/* TODO: what to do */
			goto logoutandbail;
		}
		/* call to pop_header initilizes global vars. Now place in display */
		do_pop_process(mtypel);
	}

    if (mtypel == MONNEW) {
		show_display(TRUE);
	}
	else {
		M_DEBUG(FPRINTF(stderr,"do_mailPOP: make_display()\n"));
		make_display(mtypel);         /* Output icons and tidyup chain of records. */
	}

logoutandbail:
	if (mailItemArray) {
		free(mailItemArray);
	}
	ret = pop_close_connection(popdata);
	ret = free_popdata(popdata);
	return;
}

/* called for each line of the LIST of messages (see pop_fetch_data above) */
int 
collect_list(char* pline, void* data)
{
	long lmsgnum, lmsgsize;

	sscanf(pline, "%ld %ld", &lmsgnum, &lmsgsize);
	mailItemArray[mIAIndex++] = lmsgnum;
	return 0;
}

/* called for each line of the mail header (see pop_fetch_data above) */
/* We can handle the lines like we're scanning the spool file and
 * use the scanning routines from the spool file scanning above */
int
pop_header(char* pline, void* data)
{
	enum mon_type mtypel = (enum mon_type)data;
	char* b;
	char* l;

	if (*line == '\n') {
		m_eoh = TRUE;
	}
	if (LWSP(*pline)) {
		line = realloc(line, strlen(line)+strlen(pline)+10);
		l = line + strlen(line);
		b = pline;
		while (*b && LWSP(*b)) {
			b++;
		}
		STRCPY(l, b);
		M_DEBUG(FPRINTF(stderr, "Extend: <<%s>>\n", line));
	}
	else {
		/* process the last line */
		do_pop_process(mtypel);
		/* save this line for possible later continuation */
		line = realloc(line, strlen(pline)+10);
		STRCPY(line, pline);
	}
	return 0;
}

void
do_pop_process(enum mon_type mtypel)
{
	if (strlen(line) != 0) {
		if (EQUAL(line, "From ") && m_eoh == FALSE) {
			process_from();
			pop_processed = TRUE;
		}
		else if (is822header(line, "From") && m_eoh == FALSE) {
			process_from();
			pop_processed = TRUE;
		}
		else if (is822header(line, "Date") && m_eoh == FALSE) {
			pop_this_date = parsedate(line);
		}
		else if (is822header(line, "X-Face") && m_eoh == FALSE) {
			process_face();
			pop_processed = TRUE;
		}
	}
	*line = '\0';
	/* check to see if anything procesed */
	if (pop_processed) {
		/* if only doing new, see if this email is newer */
		if (mtypel == MONNEW) {
			if (!haveseen((long)pop_this_date)) {
				process_info(mtypel);
				newmail = 1;
			}
			else {
				/* Reset for the next message. */
				/* We do this reset because we didn't process the
				 * fetched info.  This forgets what was not used */
				fromc_found = froms_found = x_face = 0;
			}
		}
		else {
			process_info(mtypel);
		}
		remember((long)pop_this_date);
	}
	m_eoh = FALSE;
	pop_processed = FALSE;
	return;
}
#endif /* POP */

#ifdef IMAP
/* == IMAP MAILBOX ============================================ */
static void
do_mailIMAP(enum mon_type mtypel)
{
	int ret;
	IMAP_DATA* imap_data;
	char buf[LONG_STRING];
	long startind = 1;
	long sizeThisTime, sizeLastTime;
	long boxsizeThisTime, boxsizeLastTime;

	/* start with new positions for picons */
    column = row = 0;        /* Start in top left corner of pixrect. */
	newmail = 0;
    mboxshrunk = 0;
    noicons = 0;

	/* Make sure the user has specified a server and login */
	if ((imap_machine == NULL) || (strlen(imap_machine)==0)) {
		M_DEBUG(FPRINTF(stderr, "do_mailIMAP: imap_machine not set. Exitting\n"));
		return;
	}

	/* if doing a new scan, start afresh */
    if (mtypel == MONNEW) {
        make_pixrect(maxcols);
    }

	/* Connect to the IMAP server */
	imap_data = build_imapdata(imap_machine, IMAP_PORT, imap_user, imap_pwd);
	ret = imap_open_connection(imap_data);
	if (ret != IMAP_A_SUCCESS) {
		/* we need to tell the user something is wrong */
		/* TODO: error message passing and display */
		display_error("Could not connect or login to IMAP server");
		have_login_info = FALSE;
		goto imaplogoutandbail;
	}
	have_login_info = TRUE;

	/* STAT the mailbox and see if it has changed */
	/* (STAT returns the number of messages and malbox size in octets) */
	ret = imap_get_statinfo(imap_data, &sizeThisTime, &boxsizeThisTime);
	if (ret != IMAP_A_SUCCESS) {
		goto imaplogoutandbail;
	}
	M_DEBUG(FPRINTF(stderr,"do_mailIMAP: msgs=%ld, boxsize=%ld\n", 
					sizeThisTime, boxsizeThisTime));

	sizeLastTime = imap_last_messages;
	boxsizeLastTime = imap_last_boxsize;
	/* remember mailbox sizes for the next time we're called */
	imap_last_messages = sizeThisTime;
	imap_last_boxsize = boxsizeThisTime;

	/* if nothing to display */
	if (sizeThisTime == 0) {
        add_face(DISP_ICON, NOMAIL, "");
        if (mtypel == MONNEW) {
            show_display(TRUE);  /* Show new mail. */
        } else {
            make_display(mtypel);   /* Output icons and tidyup chain of records. */
        }
		goto imaplogoutandbail;
	}

	/* if the statistics of the mailbox has not changed, we don't need to rescan */
	if ((sizeLastTime == sizeThisTime) && (boxsizeLastTime == boxsizeThisTime)) {
		M_DEBUG(FPRINTF(stderr,"do_mailIMAP: no changes, show_display(FALSE)\n"));
		show_display(FALSE);
		goto imaplogoutandbail;
	}

	if (sizeThisTime > sizeLastTime) {
		M_DEBUG(FPRINTF(stderr,"do_mailIMAP: new mail to add\n"));
		newmail = 1;
	}
	else {
		/* when the mailbox shrinks. There can't be new stuff to show */
		M_DEBUG(FPRINTF(stderr,"do_mailIMAP: mailbox shrinking\n"));
		mboxshrunk = 1;
	}

	if (line) {
		FREE(line);
	}
	line = Malloc(MAXLINE);
	*line = '\0';

	/* Use  the FETCH command for the mail headers */
	startind = MAX(1,startind);		/* get rid of any first time zeros */
	if (mtypel == MONNEW) {
		/* if MONNEW, we only look through the last several entries for new ones*/
		startind = MAX(startind, imap_last_messages-maxcols-5);
	}

	m_eoh = FALSE;
	imap_processed = FALSE;
	if (startind <= sizeThisTime) {
		sprintf(buf, "FETCH %ld:%ld RFC822.HEADER.LINES (FROM DATE X-FACE)", 
				startind, sizeThisTime);
		imap_processed = FALSE;
		ret = imap_query(imap_data, buf, imap_header, (void *)mtypel);
	}
	
    if (mtypel == MONNEW) {
		show_display(TRUE);
	}
	else {
		M_DEBUG(FPRINTF(stderr,"do_mailIMAP: make_display()\n"));
		make_display(mtypel);         /* Output icons and tidyup chain of records. */
	}

imaplogoutandbail:
	ret = imap_close_connection(imap_data);
	ret = free_imapdata(imap_data);
	return;
}

/* The fetch retuns the header lines for the messages separated
 * by a CMD_RESPOND line.  We use the latter to force processing.
 */
int
imap_header(int icode, char* buf, void* data)
{
	enum mon_type mtypel = (enum mon_type)data;
	char* b;
	char* l;

	if (icode & IMAP_CMD_TAG) {
		/* if it's a tag, just process the last one if there was one */
		do_imap_process(mtypel);
		return 0;
	}
	else if (icode & IMAP_CMD_RESPOND) {
		/* the response should be the FETCH saying it's doing it */
		do_imap_process(mtypel);
		return 0;
	}
	else {
		if (*buf == '\n') {
			m_eoh = TRUE;
		}
		/* if the line starts with white space, it must be a continuation
		 * line from the previous header line.  If so, strip while space
		 * from beginning and append to the last line */
		if (LWSP(*buf)) {
			line = realloc(line, strlen(line)+strlen(buf)+10);
			l = line + strlen(line);
			b = buf;
			while (*b && LWSP(*b)) {
				b++;
			}
			STRCPY(l, b);
			M_DEBUG(FPRINTF(stderr, "Extend: <<%s>>\n", line));
		}
		else {
			/* process the last line */
			do_imap_process(mtypel);
			/* save this line for possible later continuation*/
			line = realloc(line, strlen(buf)+10);
			STRCPY(line, buf);
		}
	}

	return 0;
}

void
do_imap_process(enum mon_type mtypel)
{
	if (strlen(line) != 0) {
		if (EQUAL(line, "From ") && m_eoh == FALSE) {
			process_from();
			imap_processed = TRUE;
		}
		else if (is822header(line, "From") && m_eoh == FALSE) {
			process_from();
			imap_processed = TRUE;
		}
		else if (is822header(line, "X-Face") && m_eoh == FALSE) {
			process_face();
			imap_processed = TRUE;
		}
		else if (is822header(line, "Date") && m_eoh == FALSE) {
			imap_this_date = parsedate(line);
		}
	}
	*line = '\0';
	/* check to see if anything procesed */
	if (imap_processed) {
		/* if only doing new, see if this email is newer */
		if (mtypel == MONNEW) {
			if (!haveseen((long)imap_this_date)) {
				process_info(mtypel);
				newmail = 1;
			}
			else {
				/* Reset for the next message. */
				/* We do this reset because we didn't process the
				 * fetched info.  This forgets what was not used */
				fromc_found = froms_found = x_face = 0;
			}
		}
		else {
			process_info(mtypel);
		}
		remember((long)imap_this_date);
	}
	m_eoh = FALSE;
	imap_processed = FALSE;
	return;
}
#endif /* IMAP */

/* ============================================================= */
/* Run user supplied program or script. */
static void
do_prog(enum mon_type mtypel)
{
    FILE *fp;                /* File descriptor for users command output. */
    int cols = maxcols;      /* Maximum number of columns for display. */
    int isicon = 0;          /* Set if there is an initial icon record. */
    int rows = 1;            /* Maximum number of rows for display. */
    char *reply;

    if (userprog == NULL ||
        (fp = popen(userprog, "r")) == NULL) {   /* Connect to user program. */
        FPRINTF(stderr,"%s: couldn't get program (%s) information.\n",
                progname, userprog);
        exit(1);
    }
    if ((reply = fgets(nextline, MAXLINE, fp)) == NULL || !reply) {
        FPRINTF(stderr,"%s: couldn't get program (%s) information.\n",
                progname, userprog);
        exit(1);
    }
    if (!EQUAL(nextline, "Cols=")) {
        isicon = 1;
        get_userrec(nextline, DISP_ICON);
        FGETS(nextline, MAXLINE, fp);
    }
    SSCANF(nextline, "Cols=%d Rows=%d", &cols, &rows);
    if (cols < 1) {
        cols = maxcols;
    }
    if (rows < 1) {
        rows = 1;
    }

    width  = cols * imagewidth;
    height = rows * imageheight;
    create_pixrects(width, height);

    maxcols = cols;          /* Maximum width in columns. */
    column = row = 0;        /* Start in top left corner of pixrect. */
    while (fgets(nextline, MAXLINE, fp) != NULL) {
        get_userrec(nextline, (isicon ? DISP_BOTH : DISP_ALL));
        adjust(mtypel);            /* Adjust column and row. */
    }
    PCLOSE(fp);
    show_display(TRUE);
}

static void
get_userrec(char *s, enum disp_type dtype)
{
    enum icon_type ftype;       /* Type of the current face. */
    char *fields[MAXFIELDS];    /* Pointers to program record fields. */
    char line[MAXLINE];
    char *fptr, *lptr;
    int i;

    STRCPY(line, s);
    for (i = 0; i < MAXFIELDS; i++) {
        fields[i] = NULL;
    }
    lptr = line;
    for (i = 0; i < MAXFIELDS; i++) {
        if ((fptr = strchr(lptr, '\t')) == NULL) {
            if ((fptr = strchr(lptr, '\n')) == NULL) {
                break;
            }
        }
        fields[i] = lptr;
        lptr = fptr + 1;
        *fptr = '\0';
    }
    for (i = (int) REC_LWIN; i <= (int) REC_RICON; i++) {
        if (fields[i] != NULL && fields[i][0] == '\0') {
            fields[i] = NULL;
        }
    }

    STRCPY(community, "");
    if (!(EQUAL(fields[(int) REC_USER], "NOMAIL")  ||
          EQUAL(fields[(int) REC_USER], "NOPAPER") ||
          EQUAL(fields[(int) REC_USER], "NOPRINT") ||
          EQUAL(fields[(int) REC_USER], "NOTHING") ||
          EQUAL(fields[(int) REC_USER], "NOUSERS"))) {

/* Turn hostname into community name. */

        h_to_c(fields[(int) REC_HOST], community);
        a_to_u(community, fields[(int) REC_USER], realcomm, realname);
        found = make_iconname(facepath, realcomm, realname);
    }

    if (EQUAL(fields[(int) REC_USER], "NOMAIL")) {
        ftype = NOMAIL;
    } else if (EQUAL(fields[(int) REC_USER], "NOPAPER")) {
        ftype = NOPAPER;
    } else if (EQUAL(fields[(int) REC_USER], "NOPRINT")) {
        ftype = NOPRINT;
    } else if (EQUAL(fields[(int) REC_USER], "NOTHING")) {
        ftype = NOTHING;
    } else if (EQUAL(fields[(int) REC_USER], "NOUSERS")) {
        ftype = NOUSERS;
    } else {
        ftype = ORDINARY;
    }

    add_face(dtype, ftype, iconname);
    if (dtype == DISP_ICON || dtype == DISP_ALL) {
        if (fields[(int) REC_LWIN] != NULL) {
            text(DISP_ICON,  LEFT,  fields[(int) REC_LWIN]);
        }
        if (fields[(int) REC_RWIN] != NULL) {
            text(DISP_ICON,  RIGHT, fields[(int) REC_RWIN]);
        }
    }
    if (dtype != DISP_ICON) {
        if (fields[(int) REC_LICON] != NULL) {
            text(DISP_OTHER, LEFT,  fields[(int) REC_LICON]);
        }
        if (fields[(int) REC_RICON] != NULL) {
            text(DISP_OTHER, RIGHT, fields[(int) REC_RICON]);
        }
        if (fields[(int) REC_LWIN] != NULL) {
            text(DISP_NAME,  LEFT,  fields[(int) REC_LWIN]);
        }
        if (fields[(int) REC_RWIN] != NULL) {
            text(DISP_NAME,  RIGHT, fields[(int) REC_RWIN]);
        }
    }
}


/* Physically place a picon into the output image.
 *
 * Note that this relies on the global variables 'row' and 'column'
 *    (use adjust() to advance through the icon placement)
 */
static void
add_face(enum disp_type dtype, enum icon_type itype, char *name)
{
    unsigned short buf[256];      /* Ikon/icon image. */

    if (itype == XFACE) {
        itype = ORDINARY;
        destroy_image(itype);
        get_xface(face_buf, buf);
        load_xbm_icon(itype, buf);
    } else if (itype == ORDINARY) {
        destroy_image(itype);
        itype = get_icon(name, itype);
    }
 
    switch (dtype) {
        case DISP_ALL   : adjust_image(DISP_NAME,  itype, row, column);
                          adjust_image(DISP_OTHER, itype, row, column);
                          adjust_image(DISP_ICON,  itype, 0, 0);
                          break;

        case DISP_BOTH  : adjust_image(DISP_NAME,  itype, row, column);
                          adjust_image(DISP_OTHER, itype, row, column);
                          break;

        case DISP_ICON  : adjust_image(dtype, itype, 0, 0);
                          break;

        case DISP_NAME  :
        case DISP_OTHER : adjust_image(dtype, itype, row, column);
    }
    if (itype == NOMAIL) {
        drop_back();
    }
}


/* Advance row/column of the displayed picons */
static void
adjust(enum mon_type mtypel)
{
    if (mtypel != MONNEW) {
        if (++column == maxcols) {
            column = 0;
            row++;
        }
        return;
    }
}

/* Check if there is an audio file for this user. */
#ifdef AUDIO_SUPPORT
void
check_audio(char *iconname)
{
    struct stat sbuf;
    char filename[MAXLINE], *ptr;

    if (do_audio && !firsttime) {
        if ((ptr = strrchr(iconname, '/')) != NULL) {
            STRNCPY(filename, iconname, ptr - iconname);
            filename[ptr - iconname] = '\0';
            STRCAT(filename, "/face.au");
            if (stat(filename, &sbuf) != -1) {
                play_sound(filename);
            }
        }
    }
}
#endif /*AUDIO_SUPPORT*/

/* Make window pixrect the correct size. */
static void
make_pixrect(int count)
{
    int c, r;               /* Size in columns and rows of window display. */

    r = ((count-1) / maxcols) + 1;   /* Number of rows of faces. */
    c = maxcols;                     /* Full width display. */
    if (count <= maxcols) {
        r = 1;         /* One row. */
        c = count;     /* Of 'count' columns. */
    }

    height = r * imageheight;        /* Height of the icon display. */
    width = c * imagewidth;          /* Width of the icon display. */
    create_pixrects(width, height);
}


/* Output icons and tidyup chain of records. */
static void
make_display(enum mon_type mtypel)
{
    int count;
    struct recinfo *this, *next;

    count = noicons;         /* Number of faces to display. */
    if (!count) {
        count = 1;           /* Always one "no mail" icon. */
    }
    make_pixrect(count);

    count = 0;
    if (!noicons) {
        add_face(DISP_ALL, NOMAIL, "");
    } else {
        this = recs;
        while (this != NULL) {
            next = this->next;
            if (!this->total) {
                this = next;
                continue;
            }
            if (this->faceimage != NULL) {
                STRCPY(face_buf, (char *) this->faceimage);
                add_face(DISP_ALL, XFACE, this->iconname);
            } else {
                add_face(DISP_ALL, ORDINARY, this->iconname);
            }
            count += this->total;
            if (!dontshowno) {
                SPRINTF(nextline, "%1d", this->total);
                text(DISP_OTHER, RIGHT, nextline);
            }
            if (!dontshowtime) {
                text(DISP_OTHER, LEFT, this->ts);
            }
            if (!dontshowuser) {
                text(DISP_NAME, LEFT, this->username);
                text(DISP_ICON, LEFT, this->username);
            }
            adjust(mtypel);
            this = next;
        }
        SPRINTF(nextline, "%1d", count);
        text(DISP_ICON, RIGHT, nextline);
        garbage_collect();                 /* Remove zero length records. */
    }
    show_display(TRUE);         /* Display the latest set of faces. */
}


#ifdef AUDIO_SUPPORT
void
play_sound(char *filename)       /* Play an audio sound file. */
{
    char cmd[MAXLINE];

    if (do_audio) {
        SPRINTF(cmd, "%s %s", audio_cmd, filename);
        SYSTEM(cmd);
    }
}
#endif /*AUDIO_SUPPORT*/


static void
process_face()        /* Extract next line of face image information. */
{

/* Read in the X-Face: data. */

    char *ptr;

/*  Only take the first occurrence of X-Face: in any given mail item.
 *  This solves several of the possible problems in this area, and
 *  prevents core dumps when an X-Face: line terminates the mail file.
 */

    if (x_face) {
        return;
    }

    CLEARMEM(face_buf, 2048);

/* XXX: This &line[7] stuff is dangerous, and relies on the fact that
 *      fgetheader() pushed the colon back against the field name, which
 *      may well change. This should be fixed.
 */

    ptr = &line[7];
    while (LWSP(*ptr)) {
        ptr++;
    }
    STRNCPY(face_buf, ptr, strlen(ptr)-1);
    face_buf[strlen(ptr)-1] = '\0';
    if (uncompface(face_buf) < 0) {
        return;
    }
    x_face = 1;                       /* We have an "on-the-fly" face. */
}


/*  Both process_from() and process_face() need to be fixed to take into
 *  account the possibility that one piece of mail is from multiple people.
 *  This, however, can wait until after we change over to a different method
 *  of face lookup (full name?).
 */
 
static void
process_from()             /* Extract information from the From line. */
{
    char *host, *pfline, *user;
    char temp[MAXLINE], ts[9];

    pfline = line;
    if (EQUAL(line, "From ")) {
        SSCANF(line, "%s %s %s %s %s %s", temp, temp, temp, temp, temp, ts);
        ts[5] = '\0';
        STRCPY(face_ts, ts);
        froms_found = 1;
    } else if (is822header(line, "From")) {
        if (fromc_found) {             /* Only process first From: line. */
            FPRINTF(stderr,
                    "%s: warning: message has more than one `From:' header\n",
                    progname);
            return;
        }
        fromc_found = 1;

/*  parse822from() will return non-zero on OK. If it returns zero, there
 *  was an error and we fall back to using parsefrom().  If it was OK,
 *  we check the local-part for routing characters (% or !); if any are
 *  present, we run the local-part through parsefrom and use the result
 *  of that.  Otherwise, we're happy with what parse822from() returned.
 */

        if (parse822from(line, &user, &host)) {
            if (strchr(user, '!') == (char *) NULL &&
                strchr(user, '%') == (char *) NULL) {
                STRCPY(face_user, strlower(user));
                STRCPY(face_host, strlower(host));
                return;
            } else {
                SPRINTF(temp, "From: %s\n", user);
                pfline = temp;
            }
        }
    }

    if (parsefrom(pfline, &user, &host)) {
        STRCPY(face_user, strlower(user));
        STRCPY(face_host, strlower(host));
    } else {
        FPRINTF(stderr, "%s: couldn't parse '%s'\n", progname, pfline);
    }
}


static void
process_info(enum mon_type mtypel)          /* Process From line and face information. */
{
    struct recinfo *crec;    /* Pointer to current mail record for updating. */
    char *lcomm = realcomm;
    char *lname = realname;

    h_to_c(face_host, community);    /* Turn hostname into community name. */
    a_to_u(community, face_user, realcomm, realname);

    found = make_iconname(facepath, realcomm, realname);
    if (x_face && found == ICON_NOTFOUND) {
        SPRINTF(iconname, "%s/%s/%s/48x48x1", facepath[0], realcomm, realname);
    }

    if (disphost == TRUE && found == ICON_USER) {
        lname = realcomm;
        lcomm = realname;
    }

    if (mtypel == MONNEW) {
        if (x_face) {
#ifdef AUDIO_SUPPORT
            if (found == ICON_BOTH) {
                check_audio(iconname);
            }
#endif /*AUDIO_SUPPORT*/
            add_face(DISP_ALL, XFACE, iconname);
        } else {
            if (found == ICON_NOTFOUND) {
                add_face(DISP_ALL, NOFACE, (char *) NULL);
            } else {
#ifdef AUDIO_SUPPORT
                check_audio(iconname);
#endif /*AUDIO_SUPPORT*/
                add_face(DISP_ALL, ORDINARY, iconname);
            }
        }

        if (!dontshowtime) {
            text(DISP_OTHER, LEFT, face_ts);
        }
        if (!dontshowuser) {
            text(DISP_NAME, LEFT, lname);
            text(DISP_ICON, LEFT, lname);
        }
        adjust(mtypel);
    } else {
        if (x_face || (crec = rec_exists(lcomm, lname)) == NULL) {
            add_record(lcomm, lname, face_ts, 0);
        } else {
            STRCPY(crec->ts, face_ts);
            if (!crec->total) {
                noicons++;
            }
            crec->total++;
        }
    }
    fromc_found = froms_found = x_face = 0;  /* Reset for the next message. */
}

/* "remember" that we've seen the passed ID */
/* This routine and the next hide the ID storage system.  It can
 * be anything -- btrees, linear arrays, ...  Just choose your efficiency */
/* TODO: a better storage system that is faster at searching and inserts */
void
remember(long lID)
{
	if (idmem == NULL) {
		idmem_size = 100;
		idmem = (long*)Malloc(idmem_size*sizeof(long));
		idmem_last = 0;
	}
	if (idmem_last >= idmem_size) {
		idmem_size += 100;
		idmem = realloc(idmem, idmem_size*sizeof(long));
	}
	if (!haveseen(lID)) {
		idmem[idmem_last] = lID;
		idmem_last++;
	}
	return;
}

/* recall if we have seen a remembered ID */
/* Returns TRUE or FALSE if we've seen the ID before */
int
haveseen(long lID)
{
	int i;

	for (i=0; i<idmem_last; i++) {
		if (idmem[i] == lID) {
			return TRUE;
		}
	}
	return FALSE;
}
