
/*  $Header: /cvs/faces/faces/faces/parsefrom.c,v 1.2 2002/01/15 22:58:46 devface Exp $
 *
 *  Returns pointers to the user and host components of an address
 *  described in either a Unix "From " line or an RFC 822 "From:" line.
 *
 *  Copyright (c) Chris Maltby - Softway Ptd Ltd - June 1990.
 *
 *  Permission is given to distribute these sources, as long as the
 *  copyright messages are not removed, and no monies are exchanged. 
 *
 *  No responsibility is taken for any errors inherent either
 *  to the comments or the code of this program, but if reported
 *  to me then an attempt will be made to fix them.
 */

#include <stdio.h>
#include <string.h>
#include "faces.h"
#include "extern.h"


int
parsefrom(char *line, char **userp, char **hostp)
{
    char *p, *user, *host;

    if (strncmp((user = line), "From ", 5) != 0) {
        if (strncmp(user, "From:", 5) == 0) {
            user += 5;

            while (*user == ' ' || *user == '\t') {
                ++user;
            }

/* RFC822 - two formats
 * From: comment <address>
 * From: address (comment)
 *
 * Will also handle:
 * From: address
 */

            if ((p = strchr(user, '<')) != NULL) {
                user = ++p;
                if ((p = strchr(user, '>')) != NULL) {
                    *p = '\0';
                } else {
                    return(0);
                }
            } else if ((p = strrchr(user, '(' )) != NULL ||
                       (p = strrchr(user, '\n')) != NULL) {
                do {
                    *p = '\0';
                } while (*--p == ' ' || *p == '\t');
            }
        } else {
            return(0);
        }
    } else {
        user += 5;
        while (*user == ' ' || *user == '\t') {
            ++user;
        }
        if ((p = strchr(user, ' ')) != NULL) {
            *p = '\0';    /* discard date */
        }
    }

    host = NULL;

    /* look for route addresses: @host,addr */
    while (*user == '@') {
        host = ++user;
        if ((p = strchr(user, ',')) != NULL) {
            *p++ = '\0';
            user = p;
        }
    }

    /* look for '@' */
    while ((p = strrchr(user, '@')) != NULL) {
        *p++ = '\0';
        host = p;
    }

    /* now for '%' */
    while ((p = strrchr(user, '%')) != NULL) {
        *p++ = '\0';
        host = p;
    }

    /* now for '!' */
    while ((p = strchr(user, '!')) != NULL) {
        *p++ = '\0';
        host = user;
        user = p;
    }

    if (host == NULL) {
        host = hostname;
    }
#ifdef NODOMAINS
    else if ((p = strchr(host, '.')) != NULL) {
        *p = '\0';
    }
#endif /* NODOMAINS */
    if ((p = strchr(user, '\n')) != NULL) {
        *p = '\0';
    }
    if ((p = strchr(host, '\n')) != NULL) {
        *p = '\0';
    }
    *hostp = host;
    *userp = user;

    return(1);
}
